import sys
import wx

import wx.grid as Grid

from atrcopy import match_bit_mask, comment_bit_mask, data_bit_mask, selected_bit_mask, diff_bit_mask

import logging
log = logging.getLogger(__name__)


class ByteGridRenderer(Grid.PyGridCellRenderer):
    def __init__(self, table, editor):
        """Render data in the specified color and font and fontsize"""
        Grid.PyGridCellRenderer.__init__(self)
        self.table = table
        m = editor.machine
        self.color = m.text_color
        self.diff_color = m.diff_text_color
        self.font = m.text_font
        self.selected_background = m.highlight_color
        self.selected_brush = wx.Brush(m.highlight_color, wx.SOLID)
        self.selected_pen = wx.Pen(m.highlight_color, 1, wx.SOLID)
        self.normal_background = m.background_color
        self.normal_brush = wx.Brush(m.background_color, wx.SOLID)
        self.normal_pen = wx.Pen(m.background_color, 1, wx.SOLID)
        self.data_background = m.data_color
        self.data_brush = wx.Brush(m.data_color, wx.SOLID)
        self.cursor_background = m.background_color
        self.cursor_brush = wx.Brush(m.background_color, wx.TRANSPARENT)
        self.cursor_pen = wx.Pen(m.unfocused_cursor_color, 2, wx.SOLID)
        self.match_background = m.match_background_color
        self.match_brush = wx.Brush(m.match_background_color, wx.SOLID)
        self.match_pen = wx.Pen(m.match_background_color, 1, wx.SOLID)
        self.comment_background = m.comment_background_color
        self.comment_brush = wx.Brush(m.comment_background_color, wx.SOLID)
        self.comment_pen = wx.Pen(m.comment_background_color, 1, wx.SOLID)

    def Draw(self, grid, attr, dc, rect, row, col, isSelected):
        # Here we draw text in a grid cell using various fonts
        # and colors.  We have to set the clipping region on
        # the grid's DC, otherwise the text will spill over
        # to the next cell
        dc.SetClippingRect(rect)

        # clear the background
        dc.SetBackgroundMode(wx.SOLID)
        
        index, _ = self.table.get_index_range(row, col)
        if not self.table.is_index_valid(index):
            dc.SetBrush(wx.Brush(wx.WHITE, wx.SOLID))
            dc.SetPen(wx.Pen(wx.WHITE, 1, wx.SOLID))
            dc.DrawRectangleRect(rect)
        else:
            text, style = self.table.get_value_style(row, col)
            style = self.table.get_style_override(row, col, style)
            if style & selected_bit_mask:
                dc.SetPen(self.selected_pen)
                dc.SetBrush(self.selected_brush)
                dc.SetTextBackground(self.selected_background)
            elif style & match_bit_mask:
                dc.SetPen(self.match_pen)
                dc.SetBrush(self.match_brush)
                dc.SetTextBackground(self.match_background)
            elif style & comment_bit_mask:
                dc.SetPen(self.comment_pen)
                dc.SetBrush(self.comment_brush)
                dc.SetTextBackground(self.comment_background)
            elif style & data_bit_mask:
                dc.SetPen(self.normal_pen)
                dc.SetBrush(self.data_brush)
                dc.SetTextBackground(self.data_background)
            else:
                dc.SetPen(self.normal_pen)
                dc.SetBrush(self.normal_brush)
                dc.SetTextBackground(self.normal_background)
            dc.DrawRectangleRect(rect)

            dc.SetBackgroundMode(wx.SOLID)

            if style & diff_bit_mask:
                dc.SetTextForeground(self.diff_color)
            else:
                dc.SetTextForeground(self.color)
            dc.SetFont(self.font)
            dc.DrawText(text, rect.x+1, rect.y+1)

            # Okay, now for the advanced class :)
            # Let's add three dots "..."
            # to indicate that that there is more text to be read
            # when the text is larger than the grid cell

            width, height = dc.GetTextExtent(text)
            
            if width > rect.width-2:
                width, height = dc.GetTextExtent("...")
                x = rect.x+1 + rect.width-2 - width
                dc.DrawRectangle(x, rect.y+1, width+1, height)
                dc.DrawText("...", x, rect.y+1)
            
            r, c = self.table.get_row_col(grid.editor.cursor_index)
            if row == r and col == c:
                dc.SetPen(self.cursor_pen)
                dc.SetBrush(self.cursor_brush)
                x = rect.x+1
                if sys.platform == "darwin":
                    w = rect.width - 2
                    h = rect.height - 2
                else:
                    w = rect.width - 1
                    h = rect.height - 1
                dc.DrawRectangle(rect.x+1, rect.y+1, w, h)

        dc.DestroyClippingRegion()


class ByteGridTable(Grid.PyGridTableBase):
    column_labels = [""]
    column_sizes = [4]
    
    @classmethod
    def update_preferences(cls, prefs):
        if prefs.hex_grid_lower_case:
            cls.get_value_style = cls.get_value_style_lower
        else:
            cls.get_value_style = cls.get_value_style_upper
    
    def __init__(self):
        Grid.PyGridTableBase.__init__(self)
        
        self._rows = 1
        self._cols = len(self.column_labels)
    
    def get_index_range(self, row, col):
        """Get the byte offset from start of file given row, col
        position.
        """
        index = row * self.bytes_per_row + col
        return index, index

    def get_row_col(self, index):
        return divmod(index, self.bytes_per_row)

    def is_index_valid(self, index):
        return index < self._rows * self._cols
    
    def get_col_size(self, c):
        return self.column_sizes[c]
    
    def get_col_type(self, c):
        return "hex"

    def get_next_cursor_pos(self, row, col):
        col += 1
        if col >= self._cols:
            if row < self._rows - 1:
                row += 1
                col = 0
            else:
                col = self._cols - 1
        return (row, col)
    
    def get_next_editable_pos(self, row, col):
        return self.get_next_cursor_pos(row, col)
   
    def get_prev_cursor_pos(self, row, col):
        col -= 1
        if col < 0:
            if row > 0:
                row -= 1
                col = self._cols - 1
            else:
                col = 0
        return (row, col)
   
    def get_page_index(self, index, segment_page_size, dir, grid):
        if segment_page_size < 0:
            r = grid.get_num_visible_rows()
            page_size = r * self.GetNumberCols()
        else:
            page_size = segment_page_size
        return index + (dir * page_size)

    def GetNumberRows(self):
        return self._rows

    def get_label_at_index(self, index):
        return "%s" % index

    def GetRowLabelValue(self, row):
        raise NotImplementedError

    def GetNumberCols(self):
        return self._cols

    def GetColLabelValue(self, col):
        return self.column_labels[col]
    
    def get_value_style_upper(self, row, col):
        raise NotImplementedError
    
    def get_value_style_lower(self, row, col):
        raise NotImplementedError
    
    get_value_style = get_value_style_lower
    
    def get_style_override(self, row, col, style):
        """Allow subclasses to change the style"""
        return style
    
    def GetValue(self, row, col):
        index, _ = self.get_index_range(row, col)
        if self.is_index_valid(index):
            return self.get_value_style(row, col)[0]
        return ""

    def SetValue(self, row, col, value):
        raise NotImplementedError

    def ResetViewProcessArgs(self, grid, *args):
        pass
    
    def set_grid_cell_attr(self, grid, col, attr):
        attr.SetFont(grid.editor.machine.text_font)
        attr.SetBackgroundColour("white")
        renderer = grid.get_grid_cell_renderer(self, grid.editor)
        attr.SetRenderer(renderer)
    
    def set_col_attr(self, grid, col, char_width):
        attr = Grid.GridCellAttr()
        self.set_grid_cell_attr(grid, col, attr)
        log.debug("hexcol %d width=%d" % (col, char_width))
        grid.SetColMinimalWidth(col, 0)
        grid.SetColSize(col, (char_width * self.get_col_size(col)) + 4)
        grid.SetColAttr(col, attr)

    def ResetView(self, grid, *args):
        """
        (Grid) -> Reset the grid view.   Call this to
        update the grid if rows and columns have been added or deleted
        """
        oldrows=self._rows
        oldcols=self._cols
        self.ResetViewProcessArgs(grid, *args)
        log.debug("resetting view for %s" % grid)
        
        grid.BeginBatch()

        for current, new, delmsg, addmsg in [
            (oldrows, self._rows, Grid.GRIDTABLE_NOTIFY_ROWS_DELETED, Grid.GRIDTABLE_NOTIFY_ROWS_APPENDED),
            (oldcols, self._cols, Grid.GRIDTABLE_NOTIFY_COLS_DELETED, Grid.GRIDTABLE_NOTIFY_COLS_APPENDED),
        ]:

            if new < current:
                msg = Grid.GridTableMessage(self,delmsg,new,current-new)
                grid.ProcessTableMessage(msg)
            elif new > current:
                msg = Grid.GridTableMessage(self,addmsg,new-current)
                grid.ProcessTableMessage(msg)
                self.UpdateValues(grid)
        grid.EndBatch()

        # update the scrollbars and the displayed part of the grid
        dc = wx.MemoryDC()
        dc.SetFont(grid.editor.machine.text_font)
        (width, height) = dc.GetTextExtent("M")
        grid.SetDefaultRowSize(height)
        grid.SetColMinimalAcceptableWidth(width)
        grid.SetRowMinimalAcceptableHeight(height + 1)

        for col in range(self._cols):
            # Can't share GridCellAttrs among columns; causes crash when
            # freeing them.  So, have to individually allocate the attrs for
            # each column
            self.set_col_attr(grid, col, width)

        self._rows = self.GetNumberRows()
        self._cols = self.GetNumberCols()
        
        label_font = grid.editor.machine.text_font.Bold()
        grid.SetLabelFont(label_font)
        dc.SetFont(label_font)
        (width, height) = dc.GetTextExtent("M")
        grid.SetColLabelSize(height + 4)
        text = self.GetRowLabelValue(self._rows - 1)
        grid.SetRowLabelSize(width * len(text) + 4)
        
        grid.AdjustScrollbars()
        grid.ForceRefresh()

    def UpdateValues(self, grid):
        """Update all displayed values"""
        # This sends an event to the grid table to update all of the values
        msg = Grid.GridTableMessage(self, Grid.GRIDTABLE_REQUEST_VIEW_GET_VALUES)
        grid.ProcessTableMessage(msg)


class HexDigitMixin(object):
    keypad=[ wx.WXK_NUMPAD0, wx.WXK_NUMPAD1, wx.WXK_NUMPAD2, wx.WXK_NUMPAD3, 
             wx.WXK_NUMPAD4, wx.WXK_NUMPAD5, wx.WXK_NUMPAD6, wx.WXK_NUMPAD7, 
             wx.WXK_NUMPAD8, wx.WXK_NUMPAD9
             ]
    
    def isValidHexDigit(self,key):
        return key in HexDigitMixin.keypad or (key>=ord('0') and key<=ord('9')) or (key>=ord('A') and key<=ord('F')) or (key>=ord('a') and key<=ord('f'))

    def getValidHexDigit(self,key):
        if key in HexDigitMixin.keypad:
            return chr(ord('0') + key - wx.WXK_NUMPAD0)
        elif (key>=ord('0') and key<=ord('9')) or (key>=ord('A') and key<=ord('F')) or (key>=ord('a') and key<=ord('f')):
            return chr(key)
        else:
            return None


class HexTextCtrl(wx.TextCtrl,HexDigitMixin):
    def __init__(self,parent,id,parentgrid):
        # Don't use the validator here, because apparently we can't
        # reset the validator based on the columns.  We have to do the
        # validation ourselves using EVT_KEY_DOWN.
        wx.TextCtrl.__init__(self,parent, id,
                             style=wx.TE_PROCESS_TAB|wx.TE_PROCESS_ENTER|wx.NO_BORDER)
        log.debug("parent=%s" % parent)
        self.SetInsertionPoint(0)
        self.Bind(wx.EVT_TEXT, self.on_text)
        self.Bind(wx.EVT_KEY_DOWN, self.on_key_down)
        self.parentgrid=parentgrid
        self.setMode('hex')
        self.startValue=None

    def setMode(self, mode):
        self.mode=mode
        if mode=='hex':
            self.SetMaxLength(2)
            self.autoadvance=2
        elif mode=='char':
            self.SetMaxLength(1)
            self.autoadvance=1
        else:
            self.SetMaxLength(0)
            self.autoadvance=0
        self.userpressed=False

    def editingNewCell(self, value, mode='hex'):
        """
        Begin editing a new cell by determining the edit mode and
        setting the initial value.
        """
        # Set the mode before setting the value, otherwise OnText gets
        # triggered before self.userpressed is set to false.  When
        # operating in char mode (i.e. autoadvance=1), this causes the
        # editor to skip every other cell.
        self.setMode(mode)
        self.startValue=value
        self.SetValue(value)
        self.SetFocus()
        self.SetInsertionPoint(0)
        self.SetSelection(-1, -1) # select the text

    def insertFirstKey(self, key):
        """
        Check for a valid initial keystroke, and insert it into the
        text ctrl if it is one.

        @param key: keystroke
        @type key: int

        @returns: True if keystroke was valid, False if not.
        """
        ch=None
        if self.mode=='hex':
            ch=self.getValidHexDigit(key)
        elif key>=wx.WXK_SPACE and key<=255:
            ch=chr(key)

        if ch is not None:
            # set self.userpressed before SetValue, because it appears
            # that the OnText callback happens immediately and the
            # keystroke won't be flagged as one that the user caused.
            self.userpressed=True
            self.SetValue(ch)
            self.SetInsertionPointEnd()
            return True

        return False

    def on_key_down(self, evt):
        """
        Keyboard handler to process command keys before they are
        inserted.  Tabs, arrows, ESC, return, etc. should be handled
        here.  If the key is to be processed normally, evt.Skip must
        be called.  Otherwise, the event is eaten here.

        @param evt: key event to process
        """
        log.debug("key down before evt=%s" % evt.GetKeyCode())
        key=evt.GetKeyCode()
        
        if key==wx.WXK_TAB:
            wx.CallAfter(self.parentgrid.advance_cursor)
            return
        elif self.mode=='hex':
            if self.isValidHexDigit(key):
                self.userpressed=True
        elif self.mode!='hex':
            self.userpressed=True
        evt.Skip()

    def cancel_edit(self):
        log.debug("cancelling edit in hex cell editor!")
        self.SetValue(self.startValue)
        self.parentgrid.cancel_edit()
        
    def on_text(self, evt):
        """
        Callback used to automatically advance to the next edit field.
        If self.autoadvance > 0, this number is used as the max number
        of characters in the field.  Once the text string hits this
        number, the field is processed and advanced to the next
        position.
        
        @param evt: CommandEvent
        """
        log.debug("evt=%s str=%s cursor=%d" % (evt,evt.GetString(),self.GetInsertionPoint()))
        
        # NOTE: we check that GetInsertionPoint returns 1 less than
        # the desired number because the insertion point hasn't been
        # updated yet and won't be until after this event handler
        # returns.
        if self.autoadvance and self.userpressed:
            if len(evt.GetString())>=self.autoadvance and self.GetInsertionPoint()>=self.autoadvance-1:
                # FIXME: problem here with a bunch of really quick
                # keystrokes -- the interaction with the
                # underlyingSTCChanged callback causes a cell's
                # changes to be skipped over.  Need some flag in grid
                # to see if we're editing, or to delay updates until a
                # certain period of calmness, or something.
                print "advancing after edit"
                wx.CallAfter(self.parentgrid.advance_cursor)


class HexCellEditor(Grid.PyGridCellEditor,HexDigitMixin):
    """
    Cell editor for the grid, based on GridCustEditor.py from the
    wxPython demo.
    """
    def __init__(self,grid):
        Grid.PyGridCellEditor.__init__(self)
        self.parentgrid=grid

    def Create(self, parent, id, evtHandler):
        """
        Called to create the control, which must derive from wx.Control.
        *Must Override*
        """
        log.debug("")
        self._tc = HexTextCtrl(parent, id, self.parentgrid)
        self.SetControl(self._tc)

        if evtHandler:
            self._tc.PushEventHandler(evtHandler)

    def SetSize(self, rect):
        """
        Called to position/size the edit control within the cell rectangle.
        If you don't fill the cell (the rect) then be sure to override
        PaintBackground and do something meaningful there.
        """
        log.debug("rect=%s\n" % rect)
        self._tc.SetDimensions(rect.x, rect.y, rect.width+2, rect.height+2,
                               wx.SIZE_ALLOW_MINUS_ONE)

    def Show(self, show, attr):
        """
        Show or hide the edit control.  You can use the attr (if not None)
        to set colours or fonts for the control.
        """
        log.debug("show=%s, attr=%s" % (show, attr))
        Grid.PyGridCellEditor.Show(self, show, attr)

    def PaintBackground(self, dc, rect, attr):
        """
        Draws the part of the cell not occupied by the edit control.  The
        base  class version just fills it with background colour from the
        attribute.  In this class the edit control fills the whole cell so
        don't do anything at all in order to reduce flicker.
        """
        log.debug("MyCellEditor: PaintBackground\n")

    def BeginEdit(self, row, col, grid):
        """
        Fetch the value from the table and prepare the edit control
        to begin editing.  Set the focus to the edit control.
        *Must Override*
        """
        grid.editor.select_none(False)
        self.startValue = grid.GetTable().GetValue(row, col)
        mode = self.parentgrid.table.get_col_type(col)
        log.debug("row,col=(%d,%d), mode=%s" % (row, col, mode))
        self._tc.editingNewCell(self.startValue,mode)

    def EndEdit(self, row, col, grid, old_val):
        """
        Complete the editing of the current cell. Returns True if the value
        has changed.  If necessary, the control may be destroyed.
        *Must Override*
        """
        log.debug("row,col=(%d,%d)" % (row, col))

        val = self._tc.GetValue()
        if val != self.startValue:
            changed = grid.change_value(row, col, val) # update the table
            # On error, don't advance cursor
            grid.change_is_valid = changed
        else:
            changed = False

        self.startValue = ''
        self._tc.SetValue('')
        return changed

    def Reset(self):
        """
        Reset the value in the control back to its starting value.
        *Must Override*
        """
        log.debug("")
        self._tc.SetValue(self.startValue)
        self._tc.SetInsertionPointEnd()

    def IsAcceptedKey(self, evt):
        """
        Return True to allow the given key to start editing: the base class
        version only checks that the event has no modifiers.  F2 is special
        and will always start the editor.
        """
        log.debug("keycode=%d" % (evt.GetKeyCode()))

        ## We can ask the base class to do it
        #return self.base_IsAcceptedKey(evt)

        # or do it ourselves
        return (not (evt.ControlDown() or evt.AltDown()) and
                evt.GetKeyCode() != wx.WXK_SHIFT)

    def StartingKey(self, evt):
        """
        If the editor is enabled by pressing keys on the grid, this will be
        called to let the editor do something about that first key if desired.
        """
        log.debug("keycode=%d" % evt.GetKeyCode())
        key = evt.GetKeyCode()
        if not self._tc.insertFirstKey(key):
            evt.Skip()

    def StartingClick(self):
        """
        If the editor is enabled by clicking on the cell, this method will be
        called to allow the editor to simulate the click on the control if
        needed.
        """
        log.debug("")

    def Destroy(self):
        """final cleanup"""
        log.debug("")
        Grid.PyGridCellEditor.Destroy(self)

    def Clone(self):
        """
        Create a new object which is the copy of this one
        *Must Override*
        """
        log.debug("")
        return HexCellEditor(self.parentgrid)

    
class ByteGrid(Grid.Grid):
    """
    View for editing in hexidecimal notation.
    """
    short_name = "hex"

    def __init__(self, parent, task, table, **kwargs):
        """Create the HexEdit viewer
        """
        Grid.Grid.__init__(self, parent, -1, **kwargs)
        self.task = task
        self.editor = None
        self.table = table

        # The second parameter means that the grid is to take
        # ownership of the table and will destroy it when done.
        # Otherwise you would need to keep a reference to it and call
        # its Destroy method later.
        self.SetTable(self.table, True)
        self.SetMargins(0,0)
        self.SetColMinimalAcceptableWidth(10)
        self.EnableDragGridSize(False)
        self.DisableDragRowSize()

        self.RegisterDataType(Grid.GRID_VALUE_STRING, None, None)
        e = self.get_default_cell_editor()
        self.SetDefaultEditor(e)
        self.change_is_valid = True
        self.last_change_count = 0

        self.mouse_drag_started = False
        self.select_extend_mode = False
        self.multi_select_mode = False
        self.allow_range_select = True
        self.updateUICallback = None
        self.Bind(Grid.EVT_GRID_CELL_LEFT_CLICK, self.on_left_down)
        self.Bind(Grid.EVT_GRID_CELL_LEFT_DCLICK, self.on_left_dclick)
        self.GetGridWindow().Bind(wx.EVT_MOTION, self.on_motion)
        self.GetGridWindow().Bind(wx.EVT_RIGHT_DOWN, self.on_right_down)
        self.GetGridWindow().Bind(wx.EVT_LEFT_UP, self.on_left_up)
        self.Bind(wx.EVT_KEY_DOWN, self.on_key_down)
        self.Show(True)
    
    def __repr__(self):
        return "<%s at 0x%x>" % (self.__class__.__name__, id(self))
    
    def get_grid_cell_renderer(self, table, editor):
        return ByteGridRenderer(table, editor)

    def recalc_view(self):
        editor = self.task.active_editor
        if editor is not None:
            self.editor = editor
            self.table.ResetView(self, editor)
            self.table.UpdateValues(self)
            self.goto_index(editor.cursor_index)
            self.last_change_count = editor.document.change_count
    
    def refresh_view(self):
        editor = self.task.active_editor
        if editor is not None:
            if self.editor != editor:
                self.recalc_view()
            elif self.IsShown():
                #self.ForceRefresh()
                if self.last_change_count == editor.document.change_count:
                    log.debug("skipping refresh; document change count=%d" % self.last_change_count)
                else:
                    log.debug("refreshing! document change count=%d" % self.last_change_count)
                    self.Refresh()
                    self.last_change_count = editor.document.change_count
            else:
                log.debug("skipping refresh of hidden %s" % self)
    
    def get_default_cell_editor(self):
        return HexCellEditor(self)

    def on_right_down(self, evt):
        log.debug(self.GetSelectedRows())
        r, c = self.get_rc_from_event(evt)
        actions = self.get_popup_actions(r, c)
        text, style = self.table.get_value_style(r, c)
        popup_data = {'row':r, 'col':c, 'index': self.table.get_index_range(r, c)[0], 'in_selection': style&0x80}
        if actions:
            self.editor.popup_context_menu_from_actions(self, actions, popup_data)
    
    def get_popup_actions(self, r, c):
        return []

    def on_left_up(self, evt):
        self.mouse_drag_started = False
        self.select_extend_mode = False
        self.multi_select_mode = False

    def on_left_down(self, evt):
        self.mouse_drag_started = True
        c, r = (evt.GetCol(), evt.GetRow())
        e = self.editor
        if evt.ControlDown():
            self.multi_select_mode = True
            self.select_extend_mode = False
        elif evt.ShiftDown():
            self.multi_select_mode = False
            self.select_extend_mode = True
        if self.select_extend_mode:
            index1, index2 = self.table.get_index_range(r, c)
            if index1 < e.anchor_start_index:
                e.anchor_start_index = index1
                e.cursor_index = index1
            elif index2 > e.anchor_start_index:
                e.anchor_end_index = index2
                e.cursor_index = index2 - 1
            e.anchor_initial_start_index, e.anchor_initial_end_index = e.anchor_start_index, e.anchor_end_index
            e.select_range(e.anchor_start_index, e.anchor_end_index, add=self.multi_select_mode)
        else:
            self.ClearSelection()
            index1, index2 = self.table.get_index_range(r, c)
            e.anchor_initial_start_index, e.anchor_initial_end_index = index1, index2
            e.cursor_index = index1
            e.select_range(index1, index1, add=self.multi_select_mode)
        wx.CallAfter(e.index_clicked, e.cursor_index, 0, None)

    def get_rc_from_event(self, evt):
        x, y = evt.GetPosition()
        x1, y1 = self.CalcUnscrolledPosition(x, y)
        r, c = self.XYToCell(x1, y1)
        if r < 0:
            # XYToCell fails with (-1, -1) when the mouse is not within
            # the grid of cells.  XToCol with the second param True will
            # return a valid result when it's off the edge, but there's no
            # equivalent in YToRow (at least in 3.0 classic.  In Phoenix,
            # YToRow does support that param).
            c = self.XToCol(x1, True)
            r = self.YToRow(y1)
            if r < 0:
                if y1 < 0:
                    r = 0
                else:
                    r = self.table.GetNumberRows() - 1
        return r, c
    
    def get_num_visible_rows(self):
        ux, uy = self.GetScrollPixelsPerUnit()
        sx, sy = self.GetViewStart()
        w, h = self.GetGridWindow().GetClientSize().Get()
        y = sy * uy
        r0 = self.YToRow(y)
        r1 = self.YToRow(y + h)
        if r1 < 0:
            r1 = self.table.GetNumberRows() - 1
        return r1 - r0 - 1
 
    def on_motion(self, evt):
        self.on_motion_update_status(evt)
        if not self.mouse_drag_started:
            # On windows, it's possible to get a motion event before a mouse
            # down event, so need this flag to check
            return
        e = self.editor
        if evt.LeftIsDown():
            r, c = self.get_rc_from_event(evt)
            index1, index2 = self.table.get_index_range(r, c)
            update = False
            if self.select_extend_mode:
                if index1 < e.anchor_initial_start_index:
                    e.select_range(index1, e.anchor_initial_end_index, extend=True)
                    update = True
                else:
                    e.select_range(e.anchor_initial_start_index, index2, extend=True)
                    update = True
            else:
                if e.anchor_start_index <= index1:
                    if index2 != e.anchor_end_index:
                        e.select_range(e.anchor_initial_start_index, index2, extend=self.multi_select_mode)
                        update = True
                else:
                    if index1 != e.anchor_end_index:
                        e.select_range(e.anchor_initial_end_index, index1, extend=self.multi_select_mode)
                        update = True
            if update:
                e.cursor_index = index1
                wx.CallAfter(e.index_clicked, e.cursor_index, 0, None)
    
    def on_motion_update_status(self, evt):
        x, y = self.CalcUnscrolledPosition(evt.GetPosition())
        row = self.YToRow(y)
        col = self.XToCol(x)
        index, _ = self.table.get_index_range(row, col)
        if self.table.is_index_valid(index):
            label = self.table.get_label_at_index(index)
            message = self.get_status_message_at_index(index, row, col)
            self.editor.show_status_message("%s: %s %s" % (self.short_name, label, message))
    
    def get_status_message_at_index(self, index, row, col):
        return "r=%d,c=%d" % (row, col)

    def on_key_down(self, evt):
        e = self.editor
        key = evt.GetKeyCode()
        log.debug("evt=%s, key=%s" % (evt, key))
        moved = False
        index = None
        r, c = self.GetGridCursorRow(), self.GetGridCursorCol()
        if key == wx.WXK_RETURN or key == wx.WXK_TAB:
            if evt.ControlDown():   # the edit control needs this key
                evt.Skip()
            else:
                self.DisableCellEditControl()
                if evt.ShiftDown():
                    (r, c) = self.table.get_prev_cursor_pos(r, c)
                elif self.change_is_valid:
                    (r, c) = self.table.get_next_editable_pos(r, c)
                moved = True
                self.change_is_valid = True
        elif key == wx.WXK_RIGHT:
            r, c = self.table.get_next_cursor_pos(r, c)
            moved = True
        elif key == wx.WXK_LEFT:
            r, c = self.table.get_prev_cursor_pos(r, c)
            moved = True
        elif key == wx.WXK_UP:
            r = 0 if r <= 1 else r - 1
            moved = True
        elif key == wx.WXK_DOWN:
            n = self.GetNumberRows()
            r = n - 1 if r >= n - 1 else r + 1
            moved = True
        elif key == wx.WXK_PAGEUP:
            index = self.table.get_page_index(e.cursor_index, e.segment.page_size, -1, self)
            moved = True
        elif key == wx.WXK_PAGEDOWN:
            index = self.table.get_page_index(e.cursor_index, e.segment.page_size, 1, self)
            moved = True
        else:
            evt.Skip()
        
        if moved:
            if index is None:
                index, _ = self.table.get_index_range(r, c)
            refresh_self = None if e.can_copy else self
            e.set_cursor(index, False)
            r, c = self.table.get_row_col(e.cursor_index)
            self.SetGridCursor(r, c)
            self.MakeCellVisible(r, c)
            wx.CallAfter(e.index_clicked, e.cursor_index, 0, refresh_self)
 
    def on_left_dclick(self, evt):
        self.EnableCellEditControl()
        evt.Skip()

    def cancel_edit(self):
        log.debug("cancelling edit!")
        self.DisableCellEditControl()

    def advance_cursor(self):
        self.DisableCellEditControl()
        e = self.editor
        r, c = self.table.get_row_col(e.cursor_index)
        (r, c) = self.table.get_next_editable_pos(r, c)
        self.SetGridCursor(r, c)
        index1, index2 = self.table.get_index_range(r, c)
        e.set_cursor(index1, False)
        self.EnableCellEditControl()

    def goto_index(self, index):
        row, col = self.table.get_row_col(index)
        self.SetGridCursor(row, col)
        self.MakeCellVisible(row,col)

    def select_index(self, cursor):
        self.ClearSelection()
        self.goto_index(cursor)
        self.refresh_view()
    
    def change_value(self, row, col, text):
        """Called after editor has provided a new value for a cell.
        
        Can use this to override the default handler.  Return True if the grid
        should be updated, or False if the value is invalid or the grid will
        be updated some other way.
        """
        return self.table.SetValue(row, col, text) # update the table



import os
import tempfile
import shutil

class BaseParser(object):
    mime = None
    
    # List of supported filename extensions, including the leading ".".  If
    # multiple extensions are supported, put the most common one first so that
    # the file dialog will display that as the default.
    extensions = []
    
    name = "Abstract parser"
    
    def can_parse(self, metadata):
        return metadata.mime == self.mime
    
    def is_valid_extension(self, extension):
        return extension.lower() in self.extensions
    
    def get_pretty_extension_list(self):
        return ", ".join(self.extensions)
    
    def get_file_dialog_wildcard(self):
        # Using only the first extension
        wildcards = []
        if self.extensions:
            ext = self.extensions[0]
            wildcards.append("%s (*%s)|*%s" % (self.name, ext, ext))
        return "|".join(wildcards)
    
    def parse(self, guess):
        raise NotImplementedError



#------------------------------------------------------------------------------
# Copyright (c) 2007, Riverbank Computing Limited
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD license.
# However, when used with the GPL version of PyQt the additional terms described in the PyQt GPL exception also apply

#
# Author: Riverbank Computing Limited
# Description: <Enthought pyface package component>
#------------------------------------------------------------------------------


# Enthought library imports.
from traits.api import Any, HasTraits, provides

# Local imports.
from pyface.i_widget import IWidget, MWidget


@provides(IWidget)
class Widget(MWidget, HasTraits):
    """ The toolkit specific implementation of a Widget.  See the IWidget
    interface for the API documentation.
    """


    #### 'IWidget' interface ##################################################

    control = Any

    parent = Any

    ###########################################################################
    # 'IWidget' interface.
    ###########################################################################

    def destroy(self):
        if self.control is not None:
            self.control.hide()
            self.control.deleteLater()
            self.control = None

#### EOF ######################################################################



from segmentutil import DefaultSegment, AnticFontSegment, SegmentData
from omnivore.tasks.map_edit.pane_layout import task_id_with_pane_layout as map_edit_task_id
from omnivore.tasks.jumpman.pane_layout import task_id_with_pane_layout as jumpman_task_id

getaway_defaults = {
    'antic_color_registers': [0x46, 0xD6, 0x74, 0x0C, 0x14, 0x86, 0x02, 0xB6, 0xBA],
    'tile map': [
        ("road", [0x70]),
        ("trees", range(0x80, 0x96), range(0x01, 0x16),),
        ("buildings", range(0x96, 0x9f), range(0x16, 0x1f), range(0x41, 0x51), range(0x5d, 0x60),),
        ("people", range(0xf1, 0xf4), range(0x71, 0x74)),
        ("water", range(0x2e, 0x41),),
        ("bridges", range(0x69, 0x6d),),
        ("vehicles", range(0x51, 0x59),),
        ("airport", range(0x60, 0x68), [0x5f], range(0x59, 0x5d), range(0xd9, 0xdd)), 
        ("golf", range(0xa9, 0xae),),
        ("other", [0x20, 0x25, 0x26, ]),
        ("special", range(0x21, 0x25), range(0x74, 0x76),), 
        ]
    }

def Getaway(doc):
    state = doc.bytes[0:6] == [0xff, 0xff, 0x80, 0x2a, 0xff, 0x8a]
    if state.all():
        print "Found getaway.xex!!!"
        r = doc.segments[0].rawdata
        font_segment = AnticFontSegment(r[0x086:0x486], 0x2b00, name="Playfield font")
        #doc.add_user_segment(font_segment)
        segment = DefaultSegment(r[0x2086:0x6086], 0x4b00, name="Playfield map")
        segment.map_width = 256
        #doc.add_user_segment(segment)
        extra_metadata = {
            'font': (font_segment.antic_font, "Antic 5"),
            'user segments': [font_segment, segment],
            'initial segment': segment,
            }
        extra_metadata.update(getaway_defaults)
        doc.last_task_id = map_edit_task_id
        return extra_metadata
    
    state = doc.bytes[0x10:0x19] == [0x00, 0xc1, 0x80, 0x0f, 0xcc, 0x22, 0x18, 0x60, 0x0e]
    if state.all():
        print "Found getaway.atr!!!"
        r = doc.segments[0].rawdata
        font_segment = AnticFontSegment(r[0x090:0x490], 0x2b00, name="Playfield font")
        #doc.add_user_segment(font_segment)
        segment = DefaultSegment(r[0x2090:0x6090], 0x4b00, name="Playfield map")
        segment.map_width = 256
        #doc.add_user_segment(segment)
        extra_metadata = {
            'font': (font_segment.antic_font, "Antic 5"),
            'user segments': [font_segment, segment],
            'initial segment': segment,
            }
        extra_metadata.update(getaway_defaults)
        doc.last_task_id = map_edit_task_id
        return extra_metadata

def JumpmanLevelBuilder(doc):
    state = doc.bytes[0:5] == [0x96, 0x02 , 0xd0 , 0x05 , 0x80]
    if not state.all():
        return
    # Check invariant bytes in the level data to make sure
    s = doc.bytes[0x0196:0x0296]
    if s[0x3f] == 0x4c and s[0x48] == 0x20 and s[0x4b] == 0x60 and s[0x4c] == 0xff:
        print "Found jumpman level builder!!!"
        r = doc.segments[0].rawdata
        found_level = False
        user_segments = []
        for s in doc.segments:
            if s.start_addr == 0x2800 and len(s) == 0x800:
                found_level = True
        if not found_level:
            user_segments.append(DefaultSegment(r[0x0196:0x0996], 0x2800, name="Jumpman Level Data"))
            user_segments.append(DefaultSegment(r[2458:3994], 0x0a00, name="Code at $0a00"))
            user_segments.append(DefaultSegment(r[3998:6046], 0x2000, name="Code at $2000"))
            user_segments.append(DefaultSegment(r[6050:22434], 0x3000, name="Code at $3000"))
        extra_metadata = {
            'user segments': user_segments,
            'initial segment': user_segments[0],
            }
        doc.last_task_id = jumpman_task_id
        return extra_metadata

level_names = [
    "01: easy does it",
    "02: robots I",
    "03: bombs away",
    "04: jumping blocks",
    "05: vampire",
    "06: invasion",
    "07: GP I",
    "08: builder",
    "09: look out below",
    "10: hotfoot",
    "11: runaway",
    "12: robots II",
    "13: hailstones",
    "14: dragonslayer",
    "15: GP II",
    "16: ride around",
    "17: roost",
    "18: roll me over",
    "19: ladder challenge",
    "20: figureit",
    "21: jump n run",
    "22: freeze",
    "23: follow the leader",
    "24: the jungle",
    "25a: mystery maze #1",
    "25b: mystery maze #2",
    "25c: mystery maze #3",
    "26: gunfighter",
    "27: robots III",
    "28: now you see it",
    "29: going down",
    "30: GP III",
]

def JumpmanFullAtr(doc):
    state = doc.bytes[0:5] == [0x96, 0x02 , 0x80 , 0x16 , 0x80]
    if not state.all():
        return
    # Check invariant bytes in the level data to make sure
    s = doc.bytes[0x0810:0x0910]
    if s[0x3f] == 0x4c and s[0x48] == 0x20 and s[0x4b] == 0x60 and s[0x4c] == 0xff:
        print "Found jumpman ATR!!!"
        r = doc.segments[0].rawdata
        found_level = False
        user_segments = []
        start = 0x0810
        for i in range(32):
            user_segments.append(DefaultSegment(r[start:start+0x800], 0x2800, name=level_names[i]))
            start += 0x800
        user_segments.append(DefaultSegment(r[70032:71568], 0x0a00, name="Code"))
        user_segments.append(DefaultSegment(r[71568:92048], 0x2000, name="Code"))
        extra_metadata = {
            'user segments': user_segments,
            'initial segment': user_segments[0],
            }
        doc.last_task_id = jumpman_task_id
        return extra_metadata

def check_builtin(doc):
    for match in [Getaway, JumpmanLevelBuilder, JumpmanFullAtr]:
        e = match(doc)
        if e is not None:
            return e
    return dict()



""" Action definitions for Jumpman Task

"""
import os
import sys

import wx

# Enthought library imports.
from pyface.tasks.action.api import EditorAction

from omnivore.utils.wx.dialogs import prompt_for_hex
from omnivore.framework.actions import SelectAllAction, SelectNoneAction, SelectInvertAction
from omnivore.utils.jumpman import DrawObjectBounds

from commands import *


class ClearTriggerAction(EditorAction):
    name = "Clear Trigger Function"

    picked = None

    def perform(self, event):
        self.picked.trigger_function = None

class TriggerAction(EditorAction):
    name = "Set Trigger Function..."

    picked = None

    def perform(self, event):
        e = self.active_editor
        addr, error = prompt_for_hex(e.window.control, "Enter trigger subroutine address: (default hex; prefix with # for decimal)", "Function to be Activated", self.picked.trigger_function, return_error=True, default_base="hex")
        if addr is not None:
            self.picked.trigger_function = addr
            e.bitmap.save_changes()

class SelectAllJumpmanAction(EditorAction):
    name = 'Select All'
    accelerator = 'Ctrl+A'
    tooltip = 'Select the entire document'
    enabled_name = 'can_select_objects'

    def perform(self, event):
        event.task.active_editor.select_all()

class SelectNoneJumpmanAction(EditorAction):
    name = 'Select None'
    accelerator = 'Shift+Ctrl+A'
    tooltip = 'Clear selection'
    enabled_name = 'can_select_objects'

    def perform(self, event):
        event.task.active_editor.select_none()

class SelectInvertJumpmanAction(EditorAction):
    name = 'Invert Selection'
    tooltip = 'Invert selection'
    enabled_name = 'can_select_objects'

    def perform(self, event):
        event.task.active_editor.select_invert()

class FlipVerticalAction(EditorAction):
    name = "Flip Selection Vertically"
    enabled_name = 'can_copy'
    picked = None
    command = FlipVerticalCommand

    def permute_object(self, obj, bounds):
        obj.flip_vertical(bounds)

    def perform(self, event):
        e = self.active_editor
        objects = e.bitmap.mouse_mode.objects
        bounds = DrawObjectBounds.get_bounds(objects)
        print "mirroring objects: %s" % objects, "bounds", bounds
        for o in e.bitmap.mouse_mode.objects:
            self.permute_object(o, bounds)
        e.bitmap.save_changes(self.command)
        e.bitmap.mouse_mode.resync_objects()


class FlipHorizontalAction(FlipVerticalAction):
    name = "Flip Selection Horizontally"
    command = FlipHorizontalCommand

    def permute_object(self, obj, bounds):
        obj.flip_horizontal(bounds)



# Standard library imports.
import sys
import os

# Major package imports.
import numpy as np
from atrcopy import SegmentData, DefaultSegment, selected_bit_mask, comment_bit_mask, data_bit_mask, match_bit_mask

# Local imports.
from omnivore.tasks.bitmap_edit.bitmap_editor import SelectMode
from omnivore.utils.command import Overlay
from omnivore.utils.jumpman import *
from omnivore.tasks.hex_edit.commands import ChangeByteCommand, SetValueCommand

from actions import *

import logging
log = logging.getLogger(__name__)


class JumpmanSelectMode(SelectMode):
    can_paste = False

    def __init__(self, *args, **kwargs):
        SelectMode.__init__(self, *args, **kwargs)
        self.mouse_down = (0, 0)
        self.objects = []

    def cleanup(self):
        pass

    def resync_objects(self):
        pass

    def get_image_override(self):
        """ Replace the entire bit image generation in JumpmanLevelView """
        return None

    def draw_extra_objects(self, lever_builder, screen, current_segment):
        return

    def draw_overlay(self, bitimage):
        return

    def get_harvest_offset(self):
        source = self.canvas.editor.segment
        if len(source) < 0x47:
            hx = hy = 0, 0
        else:
            hx = source[0x46]
            hy = source[0x47]
        return hx, hy

    def draw_harvest_grid(self, screen):
        hx, hy = self.get_harvest_offset()
        w = 160
        h = 88
        
        # Original (slow) algorithm to determine bad locations:
        #
        # def is_allergic(x, y, hx, hy):
        #     return (x + 0x30 + hx) & 0x1f < 7 or (2 * y + 0x20 + hy) & 0x1f < 5
        #
        # Note that in the originial 6502 code, the y coord is in player
        # coords, which is has twice the resolution of graphics 7. That's the
        # factor of two in the y part. Simplifying, the bad locations can be
        # defined in sets of 32 columns and 16 rows:
        #
        # x: 16 - hx, 16 - hx + 6 inclusive
        # y: 0 - hy/2, 0 - hy/2 + 2 inclusive
        hx = hx & 0x1f
        hy = (hy & 0x1f) / 2
        startx = (16 - hx) & 0x1f
        starty = (0 - hy) & 0xf

        # Don't know how to set multiple ranges simultaneously in numpy, so use
        # a slow python loop
        s = screen.style.reshape((h, w))
#        s[:] = 0
        for x in range(startx, startx + 8):
            x = x & 0x1f
            s[0:h:, x::32] |= comment_bit_mask
        for y in range(starty, starty + 4):
            y = y & 0xf
            s[y:h:16,:] |= comment_bit_mask
    
    def get_xy(self, evt):
        c = self.canvas
        e = c.editor
        if e is not None:
            index, bit, inside = c.event_coords_to_byte(evt)
            y, x = c.byte_to_row_col(index)
            if y < e.antic_lines:
                pick = e.pick_buffer[index]
            else:
                pick = -1
            return index, x, y, pick
        return None, None, None, None

    def display_coords(self, evt, extra=None):
        c = self.canvas
        e = c.editor
        if e is not None:
            index, x, y, pick = self.get_xy(evt)
            msg = "x=%d (0x%x) y=%d (0x%x) index=%d (0x%x) pick=%d" % (x, x, y, y, index, index, pick)
            if extra:
                msg += " " + extra
            e.task.status_bar.message = msg

    def process_left_down(self, evt):
        self.display_coords(evt)

    def process_left_up(self, evt):
        self.display_coords(evt)

    def process_left_dclick(self, evt):
        print "dclick"

    def process_mouse_motion_down(self, evt):
        self.display_coords(evt)

    def process_mouse_motion_up(self, evt):
        self.display_coords(evt)

    def get_picked(self, pick):
        return self.canvas.screen_state.get_picked(pick)

    def get_trigger_popup_actions(self, evt):
        index, x, y, pick = self.get_xy(evt)
        if pick >= 0:
            obj = self.get_picked(pick)
            if not obj.single:
                obj = None
        else:
            obj = None
        clear_trigger = ClearTriggerAction(enabled=obj is not None and obj.trigger_function is not None, picked=obj, task=self.canvas.editor.task)
        trigger_action = TriggerAction(enabled=obj is not None, picked=obj, task=self.canvas.editor.task)
        actions = [clear_trigger, trigger_action]
        return actions


class AnticDSelectMode(JumpmanSelectMode):
    icon = "select.png"
    menu_item_name = "Select"
    menu_item_tooltip = "Select regions"
    min_mouse_distance = 2
    can_paste = True

    def init_post_hook(self):
        self.pending_remove = None
        self.override_state = None

    def resync_objects(self):
        """ After a redraw when the level builder has been rebuilt from the hex
        data, the objects stored locally will no longer point to current
        jumpman objects. We need to find the current objects that match up to
        the stored objects.
        """
        self.objects = self.canvas.level_builder.find_equivalent(self.objects)

    def delete_objects(self):
        if self.objects:
            self.canvas.delete_objects(self.objects)
            self.objects = []
        self.canvas.Refresh()

    def get_image_override(self):
        if not self.objects:
            self.override_state = None
            return

        e = self.canvas.editor
        playfield = e.get_playfield_segment()  # use new, temporary playfield
        _, _, self.override_state = self.canvas.redraw_current(playfield, self.objects)

        # Draw the harvest grid if a peanut is selected
        for obj in self.objects:
            if obj.single:
                self.draw_harvest_grid(playfield)
                break
        bitimage = self.canvas.get_rendered_image(playfield)
        return bitimage

    def get_picked(self, pick):
        if self.override_state:
            state = self.override_state
        else:
            state = self.canvas.get_screen_state()
        return state.get_picked(pick)

    def highlight_pick(self, evt):
        index, x, y, pick = self.get_xy(evt)
        self.mouse_down = x, y
        if pick >= 0:
            obj = self.get_picked(pick)
            if obj in self.objects:
                if evt.ControlDown():
                    self.pending_remove = obj
                else:
                    self.pending_remove = True
            else:
                self.add_to_selection(obj, evt.ControlDown())
            self.check_tolerance = True
        else:
            # don't kill multiple selection if user clicks on empty space by
            # mistake
            if not evt.ControlDown():
                self.objects = []

    def add_to_selection(self, obj, append=False):
        obj.orig_x = obj.x
        obj.orig_y = obj.y
        if append:
            self.objects.append(obj)
        else:
            self.objects = [obj]

    def move_pick(self, evt):
        if self.objects:
            index, x, y, pick = self.get_xy(evt)
            dx = x - self.mouse_down[0]
            dy = y - self.mouse_down[1]
            if self.check_tolerance and abs(dx) + abs(dy) <  self.min_mouse_distance:
                return
            self.check_tolerance = False
            bad_move = False
            for obj in self.objects:
                print "moving", obj
                print " equiv", self.canvas.level_builder.find_equivalent_object(obj)
                obj.last_x, obj.last_y = obj.x, obj.y
                _, obj.x = divmod(obj.orig_x + dx, 160)
                obj.x &= obj.valid_x_mask
                obj.y = obj.orig_y + dy
                if obj.is_offscreen():
                    bad_move = True
            if bad_move:
                for obj in self.objects:
                    obj.x, obj.y = obj.last_x, obj.last_y
            self.pending_remove = None

    def process_left_down(self, evt):
        self.highlight_pick(evt)
        self.canvas.Refresh()
        self.display_coords(evt)

    def process_mouse_motion_down(self, evt):
        self.move_pick(evt)
        self.canvas.Refresh()
        self.display_coords(evt)

    def process_left_up(self, evt):
        if self.num_clicks == 2:
            return
        if self.pending_remove is True:
            self.objects = []
        elif self.pending_remove is not None:
            self.objects.remove(self.pending_remove)
        self.pending_remove = None
        if self.objects and not self.check_tolerance:
            self.canvas.save_changes()
        else:
            self.canvas.Refresh()
        self.display_coords(evt)

    def process_mouse_motion_up(self, evt):
        self.display_coords(evt)

    def check_trigger_pick(self, evt):
        index, x, y, pick = self.get_xy(evt)
        if pick >= 0:
            obj = self.get_picked(pick)
            if obj.single:
                self.canvas.editor.set_trigger_view(obj)

    def process_left_dclick(self, evt):
        self.check_trigger_pick(evt)
        evt.Skip()

    def delete_key_pressed(self):
        self.delete_objects()

    def backspace_key_pressed(self):
        self.delete_objects()

    def get_popup_actions(self, evt):
        return self.get_trigger_popup_actions(evt)


class DrawMode(JumpmanSelectMode):
    icon = "select.png"
    menu_item_name = "Draw"
    menu_item_tooltip = "Draw stuff"
    drawing_object = Girder
    can_paste = False

    def resync_objects(self):
        # objects here are only temporary, so no need to search
        pass

    def draw_extra_objects(self, level_builder, screen, current_segment):
        level_builder.draw_objects(screen, self.objects, current_segment)

    def create_objects(self, evt, start=False):
        c = self.canvas
        e = c.editor
        if e is None:
            return
        index, x, y, pick = self.get_xy(evt)
        if start:
            self.mouse_down = x, y
        dx = x - self.mouse_down[0]
        dy = y - self.mouse_down[1]
        obj = self.drawing_object
        if obj.vertical_only:
            sx = 0
            sy = obj.default_dy if dy > 0 else -obj.default_dy
            num = max((dy + sy - 1) / sy, 1)
        elif obj.single:
            sx = obj.default_dx
            sy = 0
            num = 1
        else:
            if abs(dx) >= abs(dy):
                sx = obj.default_dx if dx > 0 else -obj.default_dx
                num = max((abs(dx) + abs(sx) - 1) / abs(sx), 1)
                sy = dy / num
            else:
                sy = obj.default_dy if dy > 0 else -obj.default_dy
                num = max((abs(dy) + abs(sy) - 1) / abs(sy), 1)
                sx = dx / num
        screen_x = (self.mouse_down[0] - obj.default_dx/2) & obj.valid_x_mask
        screen_y = self.mouse_down[1] - obj.default_dy/2

        self.objects = []
        item = obj(-1, screen_x, screen_y, num, sx, sy)
        if not item.is_offscreen():
            self.objects.append(item)
        self.check_objects(x, y)

    def check_objects(self, x, y):
        pass

    def process_left_down(self, evt):
        self.create_objects(evt, True)
        self.canvas.Refresh()
        self.display_coords(evt)

    def process_left_up(self, evt):
        if self.num_clicks == 2:
            return
        self.canvas.save_objects(self.objects)
        self.objects = []
        self.display_coords(evt)

    def process_mouse_motion_down(self, evt):
        self.create_objects(evt)
        self.canvas.Refresh()
        self.display_coords(evt)

    def process_mouse_motion_up(self, evt):
        self.create_objects(evt, True)
        self.canvas.Refresh()
        self.display_coords(evt)

class DrawGirderMode(DrawMode):
    icon = "jumpman_girder.png"
    menu_item_name = "Draw Girder"
    menu_item_tooltip = "Draw girders"
    drawing_object = Girder

class DrawLadderMode(DrawMode):
    icon = "jumpman_ladder.png"
    menu_item_name = "Draw Ladder"
    menu_item_tooltip = "Draw ladders (vertical only)"
    drawing_object = Ladder

class DrawUpRopeMode(DrawMode):
    icon = "jumpman_uprope.png"
    menu_item_name = "Draw Up Rope"
    menu_item_tooltip = "Draw up ropes (vertical only)"
    drawing_object = UpRope

class DrawDownRopeMode(DrawMode):
    icon = "jumpman_downrope.png"
    menu_item_name = "Draw Down Rope"
    menu_item_tooltip = "Draw down ropes (vertical only)"
    drawing_object = DownRope

class EraseGirderMode(DrawMode):
    icon = "jumpman_erase_girder.png"
    menu_item_name = "Erase Girder"
    menu_item_tooltip = "Erase girders"
    drawing_object = EraseGirder
    editor_trait_for_enabled = 'can_erase_objects'

class EraseLadderMode(DrawMode):
    icon = "jumpman_erase_ladder.png"
    menu_item_name = "Erase Ladder"
    menu_item_tooltip = "Erase ladders (vertical only)"
    drawing_object = EraseLadder
    editor_trait_for_enabled = 'can_erase_objects'

class EraseRopeMode(DrawMode):
    icon = "jumpman_erase_rope.png"
    menu_item_name = "Erase Rope"
    menu_item_tooltip = "Erase ropes (vertical only)"
    drawing_object = EraseRope
    editor_trait_for_enabled = 'can_erase_objects'

class DrawPeanutMode(DrawMode):
    icon = "jumpman_peanut.png"
    menu_item_name = "Draw Peanuts"
    menu_item_tooltip = "Draw peanuts (single only)"
    drawing_object = Peanut

    def __init__(self, *args, **kwargs):
        DrawMode.__init__(self, *args, **kwargs)
        self.is_bad_location = False
        self.batch = None

    def cleanup(self):
        self.canvas.editor.screen.style[:] = 0

    def get_cursor(self):
        if self.is_bad_location:
            return wx.StockCursor(wx.CURSOR_HAND)
        else:
            return wx.StockCursor(wx.CURSOR_ARROW)

    def check_objects(self, x, y):
        hx, hy = self.get_harvest_offset()
        if self.objects:
            self.is_bad_location = self.objects[0].is_bad_location(hx, hy)
        else:
            self.is_bad_location = is_bad_harvest_position(x, y, hx, hy)

    def draw_extra_objects(self, level_builder, screen, current_segment):
        if self.objects:
            obj = self.objects[0]
            obj.error = self.is_bad_location
            if not obj.error:
                hx, hy = self.get_harvest_offset()
                grid = obj.harvest_checksum(hx, hy)
                obj.error = grid in level_builder.harvest_offset_seen
        level_builder.draw_objects(screen, self.objects, current_segment)
        self.draw_harvest_grid(screen)

    def change_harvest_offset(self, evt, start=False):
        c = self.canvas
        e = c.editor
        if e is None:
            return
        index, x, y, pick = self.get_xy(evt)
        if start:
            self.batch = Overlay()
            hx, hy = self.get_harvest_offset()
            self.mouse_down = hx + x, hy + y
        else:
            dx = (self.mouse_down[0] - x) & 0x1f
            dy = ((self.mouse_down[1] - y) & 0xf) * 2
            self.display_coords(evt)
            values = [dx, dy]
            source = self.canvas.editor.segment
            cmd = ChangeByteCommand(source, 0x46, 0x48, values)
            e.process_command(cmd, self.batch)

    def process_left_down(self, evt):
        self.create_objects(evt, True)
        if self.is_bad_location:
            self.change_harvest_offset(evt, True)
            self.objects = []
        else:
            self.canvas.Refresh()
        self.display_coords(evt)

    def process_left_up(self, evt):
        if self.num_clicks == 2:
            return
        if self.batch is not None:
            c = self.canvas
            e = c.editor
            if e is None:
                return
            e.end_batch()
            self.batch = None

            # Force updating of the hex view
            e.document.change_count += 1
            e.refresh_panes()
        else:
            DrawMode.process_left_up(self, evt)
            return
        self.display_coords(evt)

    def process_mouse_motion_down(self, evt):
        if self.batch is not None:
            self.change_harvest_offset(evt)
        else:
            self.create_objects(evt)
            self.canvas.Refresh()
        self.display_coords(evt)

    def process_mouse_motion_up(self, evt):
        self.create_objects(evt, True)
        self.canvas.Refresh()
        self.display_coords(evt)

    def get_popup_actions(self, evt):
        return self.get_trigger_popup_actions(evt)


class JumpmanRespawnMode(DrawMode):
    icon = "jumpman_respawn.png"
    menu_item_name = "Set Jumpman Start"
    menu_item_tooltip = "Set jumpman respawn position"
    drawing_object = JumpmanRespawn

    def init_post_hook(self):
        x, y = self.get_respawn_point()
        self.current = JumpmanRespawn(-1, x, y, 1, 0, 0)
        self.objects = []

    def get_respawn_point(self):
        source = self.canvas.editor.segment
        x = source[0x39]
        y = source[0x3a]
        return x - 0x30, (y - 0x18) / 2

    def draw_extra_objects(self, level_builder, screen, current_segment):
        objects = [self.current]
        objects.extend(self.objects)
        level_builder.draw_objects(screen, objects, current_segment)

    def change_jumpman_respawn(self, evt):
        c = self.canvas
        e = c.editor
        if e is None:
            return
        obj = self.objects[0]
        values = [obj.x + 0x30, (obj.y * 2) + 0x18]
        e.change_bytes(0x39, 0x3b, values, "Change Respawn Point")

    def process_left_up(self, evt):
        if self.num_clicks == 2:
            return
        self.change_jumpman_respawn(evt)
        self.init_post_hook()
        self.display_coords(evt)
        self.canvas.editor.refresh_panes()

    def process_mouse_motion_down(self, evt):
        self.process_mouse_motion_up(evt)



from udis import disasm, miniasm, cputables

from atrcopy import match_bit_mask, comment_bit_mask, data_bit_mask, selected_bit_mask, user_bit_mask


class BaseDisassembler(disasm.Disassembler):
    name = "generic disassembler"
    cpu = "undefined"
    allow_undocumented = False
    read_instructions = set()
    write_instructions = set()
    rw_modes = set()
    
    cached_miniassemblers = {}
    
    def __init__(self, asm_syntax, memory_map=None, hex_lower=True, mnemonic_lower=False):
        disasm.Disassembler.__init__(self, self.cpu, asm_syntax, memory_map, self.allow_undocumented, hex_lower, mnemonic_lower, self.read_instructions, self.write_instructions, self.rw_modes)

    @classmethod
    def get_nop(cls):
        cpu = cputables.processors[cls.cpu]
        return cpu['nop']
    
    def get_style(self):
        if self.pc >= self.origin + self.length:
            return 0, 0
        try:
            style = self.source.style[self.pc + self.pc_offset]
            if style & user_bit_mask:
                user_type = self.source.get_user_data(self.pc + self.pc_offset, 1)
            else:
                user_type = 0
            return style, user_type
        except AttributeError:
            # If there is no style parameter, assume it's all code!
            return 0, 0
    
    def is_data(self):
        return self.get_style()[0] & data_bit_mask
    
    def get_data_comment(self, style_tuple, bytes):
        style, user_type = style_tuple
        if (style & user_bit_mask) == 1:
            user_parser, user_disasm = user_disassemblers[user_type]
            return "; " + user_disasm(bytes, self.hex_lower)
        return disasm.Disassembler.get_data_comment(self, style, bytes)
    
    def is_same_data_style(self, style_tuple, first_style_tuple, bytes):
        style, user_type = style_tuple
        first_style, first_user_type = first_style_tuple
        if (style & first_style & user_bit_mask) == 1:
            if user_type == first_user_type:
                # display list! check to see if the byte is still part of the same
                # line in the display list disassembly
                user_parser, user_disasm = user_disassemblers[user_type]
                dl = user_parser(list(bytes))
                return len(dl) == 1
            return False
        return style & data_bit_mask and not style & comment_bit_mask
    
    @classmethod
    def get_miniassembler(cls, cpu):
        if not cpu in cls.cached_miniassemblers:
            asm = miniasm.MiniAssembler(cpu, allow_undocumented=cls.allow_undocumented)
            cls.cached_miniassemblers[cpu] = asm

        return cls.cached_miniassemblers[cpu]
    
    def assemble_text(self, pc, cmd):
        miniasm = self.get_miniassembler(self.cpu)
        bytes = miniasm.asm(pc, cmd)
        if not bytes:
            raise RuntimeError("Unknown addressing mode")
        return bytes


class Basic6502Disassembler(BaseDisassembler):
    name = "6502"
    cpu = "6502"
    read_instructions = {"adc", "and", "asl", "bit", "cmp", "cpx", "cpy", "dec", "eor", "inc", "lda", "ldx", "ldy", "lsr", "ora", "rol", "ror", "sbc", "jsr", "jmp"}
    write_instructions = {"sax", "shx", "shy", "slo", "sre", "sta", "stx", "sty"}
    rw_modes = {"absolute", "absolutex", "absolutey", "indirect", "indirectx", "indirecty", "relative", "zeropage", "zeropagex", "zeropagey"}


class Undocumented6502Disassembler(Basic6502Disassembler):
    name = "6502 (with undocumented opcodes)"
    allow_undocumented = True


class Flagged6502Disassembler(Undocumented6502Disassembler):
    name = "6502 (highlighted undocumented opcodes)"
    
    def get_flag(self, flag):
        return flag & disasm.und


class Basic65C02Disassembler(Basic6502Disassembler):
    name = "65c02"
    cpu = "65c02"


class Basic65816Disassembler(Basic6502Disassembler):
    name = "65816"
    cpu = "65816"


class Basic6800Disassembler(BaseDisassembler):
    name = "6800"
    cpu = "6800"


class Basic6809Disassembler(BaseDisassembler):
    name = "6809"
    cpu = "6809"


class Basic6811Disassembler(BaseDisassembler):
    name = "6811"
    cpu = "6811"


class Basic8051Disassembler(BaseDisassembler):
    name = "8051"
    cpu = "8051"


class Basic8080Disassembler(BaseDisassembler):
    name = "8080"
    cpu = "8080"


class BasicZ80Disassembler(BaseDisassembler):
    name = "Z80"
    cpu = "z80"


# Display list processing based on code from the atari800 project, monitor.c

def parse_antic_dl(bytes):
    groups = []
    last = []
    try:
        while len(bytes) > 0:
            byte = bytes.pop(0)
            if (byte & 0x0f == 1) or (byte & 0xf0 == 0x40):
                if last:
                    groups.append(last)
                    last = []
                # handle JVB, JMP, LMS
                last.append(byte)
                lo = bytes.pop(0)
                last.append(lo)
                hi = bytes.pop(0)
                last.append(hi)
                groups.append(last)
                last = []
            else:
                if last:
                    if last[-1] == byte:
                        last.append(byte)
                    else:
                        groups.append(last)
                        last = [byte]
                else:
                    last = [byte]
    except IndexError:
        pass
    if last:
        groups.append(last)
    return groups

def get_antic_dl(group, hex_lower=True):
    """Get the text version of the display list entry.
    
    If multiple entries are present in the group, they are assumed to be
    identical to the first entry -- i.e. they have been processed by a call to
    parse_antic_dl above.
    """
    if hex_lower:
        op_fmt = "%s %02x%02x"
    else:
        op_fmt = "%s %02X%02X"
    commands = []
    byte = group[0]
    count = len(group)
    if byte & 0xf == 1:
        if byte & 0x80:
            commands.append("DLI")
        if byte & 0x40:
            op = "JVB"
        elif byte & 0xf0 > 0:
            op = "<invalid %02x>" % byte
        else:
            op = "JMP"
        if count < 3:
            commands.append("%s <bad addr>" % op)
        else:
            commands.append(op_fmt % (op, group[2], group[1]))
    else:
        if byte & 0xf == 0:
            if byte & 0x80:
                commands.append("DLI")
            commands.append("%d BLANK" % (((byte >> 4) & 0x07) + 1))
        else:
            if byte & 0x80:
                commands.append("DLI")
            if byte & 0x40:
                if count < 3:
                    commands.append("LMS <bad addr>")
                else:
                    commands.append(op_fmt % ("LMS", group[2], group[1]))
                count = 1
            if byte & 0x20:
                commands.append("VSCROL")
            if byte & 0x10:
                commands.append("HSCROL")
            commands.append("MODE %X" % (byte & 0x0f))
        if count > 1:
            commands[0:0] = ["%dx" % count]
    return " ".join(commands)

jumpman_commands = {
    0xfc: (2, "type $ADDR"),
    0xfd: (3, "draw x=%s y=%s len=%s"),
    0xfe: (2, "spacing dx=%s dy=%s"),
    0xff: (0, "end"),
}

def parse_jumpman_level(bytes):
    groups = []
    count = 0
    data = []
    while len(bytes) > 0:
        byte = bytes.pop(0)
        if count == 0:
            if byte in jumpman_commands:
                count = jumpman_commands[byte][0]
                data = [byte]
            else:
                groups.append([byte])
                data = []
        else:
            data.append(byte)
            count -= 1

        if count == 0:
            groups.append(data)
            data = []
    if data:
        groups.append(data)
    return groups

def get_jumpman_level(group, hex_lower=True):
    """Get the text version of the display list entry.
    
    If multiple entries are present in the group, they are assumed to be
    identical to the first entry -- i.e. they have been processed by a call to
    parse_antic_dl above.
    """
    if hex_lower:
        op_fmt = "%02x"
        addr_fmt = "%04x"
    else:
        op_fmt = "%02X"
        addr_fmt = "%04X"
    commands = []
    byte = group[0]
    if byte in jumpman_commands:
        count, text = jumpman_commands[byte]
        if "ADDR" in text:
            if len(group[1:]) == 2:
                addr = addr_fmt % (group[1] + 256 * group[2])
            else:
                addr = "<bad addr>"
            text = text.replace("ADDR", addr)
        else:
            values = [op_fmt % g for g in group[1:]]
            if len(values) < count:
                values.extend(["<missing>" for i in range(count - len(values))])
            text = text % tuple(values)
    else:
        text = "<invalid command %s>" % (op_fmt % byte)
    return text

def parse_jumpman_harvest(bytes):
    groups = []
    count = 7
    data = []
    while len(bytes) > 0:
        byte = bytes.pop(0)
        if count == 7 and byte == 0xff:
            data = [byte]
            break
        data.append(byte)
        count -= 1

        if count == 0:
            groups.append(data)
            data = []
            count = 7
    if data:
        groups.append(data)
    return groups

def get_jumpman_harvest(group, hex_lower=True):
    """Get the text version of the display list entry.
    
    If multiple entries are present in the group, they are assumed to be
    identical to the first entry -- i.e. they have been processed by a call to
    parse_antic_dl above.
    """
    if hex_lower:
        op_fmt = "%02x"
        addr_fmt = "%04x"
    else:
        op_fmt = "%02X"
        addr_fmt = "%04X"
    commands = []
    byte = group[0]
    if byte == 0xff:
        text = "end"
    else:
        text = "enc=%s x=%s y=%s take=$%s paint=$%s"
        values = ["<missing>" for i in range(5)]
        if len(group) >= 1:
            values[0] = op_fmt % group[0]
        if len(group) >= 2:
            values[1] = op_fmt % group[1]
        if len(group) >= 3:
            values[2] = op_fmt % group[2]
        if len(group) >= 5:
            values[3] = addr_fmt % (group[3] + 256 * group[4])
        if len(group) >= 7:
            values[4] = addr_fmt % (group[5] + 256 * group[6])
        text = text % tuple(values)
    return text

ANTIC_DISASM = 0
JUMPMAN_LEVEL = 1
JUMPMAN_HARVEST = 2
user_disassemblers = {
    ANTIC_DISASM: (parse_antic_dl, get_antic_dl),
    JUMPMAN_LEVEL: (parse_jumpman_level, get_jumpman_level),
    JUMPMAN_HARVEST: (parse_jumpman_harvest, get_jumpman_harvest),
}



#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   12/22/2004
#
#------------------------------------------------------------------------------

""" Trait definition for a wxPython-based font.
"""

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

import wx

from traits.api \
    import Trait, TraitHandler, TraitError

#-------------------------------------------------------------------------------
#  Convert a string into a valid 'wxFont' object (if possible):
#-------------------------------------------------------------------------------

# Mapping of strings to valid wxFont families
font_families = {
    'default':    wx.DEFAULT,
    'decorative': wx.DECORATIVE,
    'roman':      wx.ROMAN,
    'script':     wx.SCRIPT,
    'swiss':      wx.SWISS,
    'modern':     wx.MODERN
}

# Mapping of strings to wxFont styles
font_styles = {
    'slant':  wx.SLANT,
    'italic': wx.ITALIC
}

# Mapping of strings wxFont weights
font_weights = {
    'light': wx.LIGHT,
    'bold':  wx.BOLD
}

# Strings to ignore in text representations of fonts
font_noise = [ 'pt', 'point', 'family' ]

#-------------------------------------------------------------------------------
#  Converts a wx.Font into a string description of itself:
#-------------------------------------------------------------------------------

def font_to_str ( font ):
    """ Converts a wx.Font into a string description of itself.
    """
    weight = { wx.LIGHT:  ' Light',
               wx.BOLD:   ' Bold'   }.get( font.GetWeight(), '' )
    style  = { wx.SLANT:  ' Slant',
               wx.ITALIC: ' Italic' }.get( font.GetStyle(), '' )
    underline = ''
    if font.GetUnderlined():
        underline = ' underline'
    return '%s point %s%s%s%s' % (
           font.GetPointSize(), font.GetFaceName(), style, weight, underline )

#-------------------------------------------------------------------------------
#  Create a TraitFont object from a string description:
#-------------------------------------------------------------------------------

def create_traitsfont ( value ):
    """ Create a TraitFont object from a string description.
    """
    if isinstance( value, wx.Font ):
        value = font_to_str( value )

    point_size = None
    family     = wx.DEFAULT
    style      = wx.NORMAL
    weight     = wx.NORMAL
    underline  = 0
    facename   = []
    for word in value.split():
        lword = word.lower()
        if font_families.has_key( lword ):
            family = font_families[ lword ]
        elif font_styles.has_key( lword ):
            style = font_styles[ lword ]
        elif font_weights.has_key( lword ):
            weight = font_weights[ lword ]
        elif lword == 'underline':
            underline = 1
        elif lword not in font_noise:
            if point_size is None:
                try:
                    point_size = int( lword )
                    continue
                except:
                    pass
            facename.append( word )
    return TraitsFont( point_size or 10, family, style, weight, underline,
                    ' '.join( facename ) )

#-------------------------------------------------------------------------------
#  'TraitsFont' class:
#-------------------------------------------------------------------------------

class TraitsFont ( wx.Font ):
    """ A Traits-specific wx.Font.
    """
    #---------------------------------------------------------------------------
    #  Returns the pickleable form of a TraitsFont object:
    #---------------------------------------------------------------------------

    def __reduce_ex__ ( self, protocol ):
        """ Returns the pickleable form of a TraitsFont object.
        """
        return ( create_traitsfont, ( font_to_str( self ), ) )

    #---------------------------------------------------------------------------
    #  Returns a printable form of the font:
    #---------------------------------------------------------------------------

    def __str__ ( self ):
        """ Returns a printable form of the font.
        """
        return font_to_str( self )

#-------------------------------------------------------------------------------
#  'TraitWXFont' class'
#-------------------------------------------------------------------------------

class TraitWXFont ( TraitHandler ):
    """ Ensures that values assigned to a trait attribute are valid font
    descriptor strings; the value actually assigned is the corresponding
    TraitsFont.
    """
    #---------------------------------------------------------------------------
    #  Validates that the value is a valid font:
    #---------------------------------------------------------------------------

    def validate ( self, object, name, value ):
        """ Validates that the value is a valid font descriptor string. If so,
        it returns the corresponding TraitsFont; otherwise, it raises a
        TraitError.
        """
        if value is None:
            return None

        try:
            return create_traitsfont( value )
        except:
            pass

        raise TraitError(object, name, 'a font descriptor string',
                         repr(value))

    def info ( self ):
        return ( "a string describing a font (e.g. '12 pt bold italic "
                 "swiss family Arial' or 'default 12')" )

#-------------------------------------------------------------------------------
#  Define a wxPython specific font trait:
#-------------------------------------------------------------------------------

### Note: Declare the editor to be a function which returns the FontEditor
# class from traits ui to avoid circular import issues. For backwards
# compatibility with previous Traits versions, the 'editors' folder in Traits
# project declares 'from api import *' in its __init__.py. The 'api' in turn
# can contain classes that have a Font trait which lead to this file getting
# imported. This leads to a circular import when declaring a Font trait.
def get_font_editor(*args, **traits):
    from font_editor import ToolkitEditorFactory
    return ToolkitEditorFactory(*args, **traits)

fh     = TraitWXFont()
WxFont = Trait( wx.SystemSettings_GetFont( wx.SYS_DEFAULT_GUI_FONT ), fh, editor = get_font_editor )



##########################################################################
#
# Processor specific code

# CPU = "6502"
# Description = "MOS Technology (and others) 6502 8-bit microprocessor."
# DataWidth = 8  # 8-bit data
# AddressWidth = 16  # 16-bit addresses

# Maximum length of an instruction (for formatting purposes)
maxLength = 3

# Leadin bytes for multibyte instructions
leadInBytes = []

# Addressing mode table
# List of addressing modes and corresponding format strings for operands.
addressModeTable = {
"implicit"    : "",
"absolute"    : "${1:02X}{0:02X}",
"absolutex"   : "${1:02X}{0:02X},x",
"absolutey"   : "${1:02X}{0:02X},y",
"accumulator" : "a",
"immediate"   : "#${0:02X}",
"indirectx"   : "(${0:02X},x)",
"indirecty"   : "(${0:02X}),y",
"indirect"    : "(${1:02X}{0:02X})",
"relative"    : "${0:04X}",
"zeropage"    : "${0:02X}",
"zeropagex"   : "${0:02X},x",
"zeropagey"   : "${0:02X},y",
}

# Op Code Table
# Key is numeric opcode (possibly multiple bytes)
# Value is a list:
#   # bytes
#   mnemonic
#   addressing mode
#   flags (e.g. pcr)
opcodeTable = {
0x00 : [ 1, "brk", "implicit"        ],
0x01 : [ 2, "ora", "indirectx"       ],
0x02 : [ 1, "hlt", "implicit", und   ],
0x03 : [ 2, "slo", "indirectx", und  ],
0x04 : [ 2, "nop", "zeropage", und   ],
0x05 : [ 2, "ora", "zeropage"        ],
0x06 : [ 2, "asl", "zeropage"        ],
0x07 : [ 2, "slo", "zeropage", und   ],
0x08 : [ 1, "php", "implicit"        ],
0x09 : [ 2, "ora", "immediate"       ],
0x0a : [ 1, "asl", "accumulator"     ],
0x0b : [ 2, "anc", "immediate", und  ],
0x0c : [ 3, "nop", "absolute", und   ],
0x0d : [ 3, "ora", "absolute"        ],
0x0e : [ 3, "asl", "absolute"        ],
0x0f : [ 3, "slo", "absolute", und   ],

0x10 : [ 2, "bpl", "relative", pcr   ],
0x11 : [ 2, "ora", "indirecty"       ],
0x12 : [ 1, "hlt", "implicit", und   ],
0x13 : [ 2, "slo", "indirecty", und  ],
0x14 : [ 2, "nop", "zeropagex", und  ],
0x15 : [ 2, "ora", "zeropagex"       ],
0x16 : [ 2, "asl", "zeropagex"       ],
0x17 : [ 2, "slo", "zeropagex", und  ],
0x18 : [ 1, "clc", "implicit"        ],
0x19 : [ 3, "ora", "absolutey"       ],
0x1a : [ 1, "nop", "implicit", und   ],
0x1b : [ 3, "slo", "absolutey", und  ],
0x1c : [ 3, "nop", "absolutex", und  ],
0x1d : [ 3, "ora", "absolutex"       ],
0x1e : [ 3, "asl", "absolutex"       ],
0x1f : [ 3, "slo", "absolutex", und  ],

0x20 : [ 3, "jsr", "absolute"        ],
0x21 : [ 2, "and", "indirectx"       ],
0x22 : [ 1, "hlt", "implicit", und   ],
0x23 : [ 2, "rla", "indirecty", und  ],
0x24 : [ 2, "bit", "zeropage"        ],
0x25 : [ 2, "and", "zeropage"        ],
0x26 : [ 2, "rol", "zeropage"        ],
0x27 : [ 2, "rla", "zeropage", und   ],
0x28 : [ 1, "plp", "implicit"        ],
0x29 : [ 2, "and", "immediate"       ],
0x2a : [ 1, "rol", "accumulator"     ],
0x2b : [ 2, "anc", "immediate", und  ],
0x2c : [ 3, "bit", "absolute"        ],
0x2d : [ 3, "and", "absolute"        ],
0x2e : [ 3, "rol", "absolute"        ],
0x2f : [ 3, "rla", "absolute", und   ],

0x30 : [ 2, "bmi", "relative", pcr   ],
0x31 : [ 2, "and", "indirecty"       ],
0x32 : [ 1, "hlt", "implicit", und   ],
0x34 : [ 2, "nop", "zeropagex", und  ],
0x35 : [ 2, "and", "zeropagex"       ],
0x36 : [ 2, "rol", "zeropagex"       ],
0x37 : [ 2, "rla", "zeropagex", und  ],
0x38 : [ 1, "sec", "implicit"        ],
0x39 : [ 3, "and", "absolutey"       ],
0x3a : [ 1, "nop", "implicit", und   ],
0x3b : [ 3, "rla", "absolutey", und  ],
0x3c : [ 3, "nop", "absolutex", und  ],
0x3d : [ 3, "and", "absolutex"       ],
0x3e : [ 3, "rol", "absolutex"       ],
0x3f : [ 3, "rla", "absolutex", und  ],

0x40 : [ 1, "rti", "implicit"        ],
0x41 : [ 2, "eor", "indirectx"       ],
0x42 : [ 1, "hlt", "implicit", und   ],
0x43 : [ 2, "sre", "indirectx", und  ],
0x44 : [ 2, "nop", "zeropage", und   ],
0x45 : [ 2, "eor", "zeropage"        ],
0x46 : [ 2, "lsr", "zeropage"        ],
0x47 : [ 2, "sre", "zeropage", und   ],
0x48 : [ 1, "pha", "implicit"        ],
0x49 : [ 2, "eor", "immediate"       ],
0x4a : [ 1, "lsr", "accumulator"     ],
0x4b : [ 2, "alr", "immediate", und  ],
0x4c : [ 3, "jmp", "absolute"        ],
0x4d : [ 3, "eor", "absolute"        ],
0x4e : [ 3, "lsr", "absolute"        ],
0x4f : [ 3, "sre", "absolute", und   ],

0x50 : [ 2, "bvc", "relative", pcr   ],
0x51 : [ 2, "eor", "indirecty"       ],
0x52 : [ 1, "hlt", "implicit", und   ],
0x53 : [ 2, "sre", "indirecty", und  ],
0x54 : [ 2, "nop", "zeropagex", und  ],
0x55 : [ 2, "eor", "zeropagex"       ],
0x56 : [ 2, "lsr", "zeropagex"       ],
0x57 : [ 2, "sre", "zeropagex", und  ],
0x58 : [ 1, "cli", "implicit"        ],
0x59 : [ 3, "eor", "absolutey"       ],
0x5a : [ 1, "nop", "implicit", und   ],
0x5b : [ 3, "sre", "absolutey", und  ],
0x5c : [ 3, "nop", "absolutex", und  ],
0x5d : [ 3, "eor", "absolutex"       ],
0x5e : [ 3, "lsr", "absolutex"       ],
0x5f : [ 3, "sre", "absolutex", und  ],

0x60 : [ 1, "rts", "implicit"        ],
0x61 : [ 2, "adc", "indirectx"       ],
0x62 : [ 1, "hlt", "implicit", und   ],
0x63 : [ 2, "rra", "indirectx", und  ],
0x64 : [ 2, "nop", "zeropage", und   ],
0x65 : [ 2, "adc", "zeropage"        ],
0x66 : [ 2, "ror", "zeropage"        ],
0x67 : [ 2, "rra", "zeropage", und   ],
0x68 : [ 1, "pla", "implicit"        ],
0x69 : [ 2, "adc", "immediate"       ],
0x6a : [ 1, "ror", "accumulator"     ],
0x6b : [ 2, "arr", "immediate", und  ],
0x6c : [ 3, "jmp", "indirect"        ],
0x6d : [ 3, "adc", "absolute"        ],
0x6e : [ 3, "ror", "absolute"        ],
0x6f : [ 3, "rra", "absolute", und   ],

0x70 : [ 2, "bvs", "relative", pcr   ],
0x71 : [ 2, "adc", "indirecty"       ],
0x72 : [ 1, "hlt", "implicit", und   ],
0x73 : [ 2, "rra", "indirecty", und  ],
0x74 : [ 2, "nop", "zeropagex", und  ],
0x75 : [ 2, "adc", "zeropagex"       ],
0x76 : [ 2, "ror", "zeropagex"       ],
0x77 : [ 2, "rra", "zeropagex", und  ],
0x78 : [ 1, "sei", "implicit"        ],
0x79 : [ 3, "adc", "absolutey"       ],
0x7a : [ 1, "nop", "implicit", und   ],
0x7b : [ 3, "rra", "absolutey", und  ],
0x7c : [ 3, "nop", "absolutex", und  ],
0x7d : [ 3, "adc", "absolutex"       ],
0x7e : [ 3, "ror", "absolutex"       ],
0x7f : [ 3, "rra", "absolutex", und  ],

0x80 : [ 2, "nop", "immediate", und  ],
0x81 : [ 2, "sta", "indirectx"       ],
0x82 : [ 2, "nop", "immediate", und  ],
0x83 : [ 2, "sax", "indirectx", und  ],
0x84 : [ 2, "sty", "zeropage"        ],
0x85 : [ 2, "sta", "zeropage"        ],
0x86 : [ 2, "stx", "zeropage"        ],
0x87 : [ 2, "sax", "zeropage", und   ],
0x88 : [ 1, "dey", "implicit"        ],
0x89 : [ 2, "nop", "immediate", und  ],
0x8a : [ 1, "txa", "implicit"        ],
0x8b : [ 2, "xaa", "immediate", und  ],
0x8c : [ 3, "sty", "absolute"        ],
0x8d : [ 3, "sta", "absolute"        ],
0x8e : [ 3, "stx", "absolute"        ],
0x8f : [ 3, "sax", "absolute", und   ],

0x90 : [ 2, "bcc", "relative", pcr   ],
0x91 : [ 2, "sta", "indirecty"       ],
0x92 : [ 1, "hlt", "implicit", und   ],
0x93 : [ 2, "sha", "indirecty", und  ],
0x94 : [ 2, "sty", "zeropagex"       ],
0x95 : [ 2, "sta", "zeropagex"       ],
0x96 : [ 2, "stx", "zeropagey"       ],
0x97 : [ 2, "sax", "zeropagey", und  ],
0x98 : [ 1, "tya", "implicit"        ],
0x99 : [ 3, "sta", "absolutey"       ],
0x9a : [ 1, "txs", "implicit"        ],
0x9b : [ 3, "shs", "absolutey", und  ],
0x9c : [ 3, "shy", "absolutex", und  ],
0x9d : [ 3, "sta", "absolutex"       ],
0x9e : [ 3, "shx", "absolutey", und  ],
0x9f : [ 3, "sha", "absolutey", und  ],

0xa0 : [ 2, "ldy", "immediate"       ],
0xa1 : [ 2, "lda", "indirectx"       ],
0xa2 : [ 2, "ldx", "immediate"       ],
0xa3 : [ 2, "lax", "indirectx", und  ],
0xa4 : [ 2, "ldy", "zeropage"        ],
0xa5 : [ 2, "lda", "zeropage"        ],
0xa6 : [ 2, "ldx", "zeropage"        ],
0xa7 : [ 2, "lax", "zeropage", und   ],
0xa8 : [ 1, "tay", "implicit"        ],
0xa9 : [ 2, "lda", "immediate"       ],
0xaa : [ 1, "tax", "implicit"        ],
0xab : [ 2, "atx", "immediate", und  ],
0xac : [ 3, "ldy", "absolute"        ],
0xad : [ 3, "lda", "absolute"        ],
0xae : [ 3, "ldx", "absolute"        ],
0xaf : [ 3, "lax", "absolute", und   ],

0xb0 : [ 2, "bcs", "relative", pcr   ],
0xb1 : [ 2, "lda", "indirecty"       ],
0xb2 : [ 1, "hlt", "implicit", und   ],
0xb3 : [ 2, "lax", "indirecty", und  ],
0xb4 : [ 2, "ldy", "zeropagex"       ],
0xb5 : [ 2, "lda", "zeropagex"       ],
0xb6 : [ 2, "ldx", "zeropagey"       ],
0xb7 : [ 2, "lax", "zeropagey", und  ],
0xb8 : [ 1, "clv", "implicit"        ],
0xb9 : [ 3, "lda", "absolutey"       ],
0xba : [ 1, "tsx", "implicit"        ],
0xbb : [ 3, "lar", "absolutey", und  ],
0xbc : [ 3, "ldy", "absolutex"       ],
0xbd : [ 3, "lda", "absolutex"       ],
0xbe : [ 3, "ldx", "absolutey"       ],
0xbf : [ 3, "lax", "absolutey", und  ],

0xc0 : [ 2, "cpy", "immediate"       ],
0xc1 : [ 2, "cmp", "indirectx"       ],
0xc2 : [ 2, "nop", "immediate", und  ],
0xc3 : [ 2, "dcp", "indirectx", und  ],
0xc4 : [ 2, "cpy", "zeropage"        ],
0xc5 : [ 2, "cmp", "zeropage"        ],
0xc6 : [ 2, "dec", "zeropage"        ],
0xc7 : [ 2, "dcp", "zeropage", und   ],
0xc8 : [ 1, "iny", "implicit"        ],
0xc9 : [ 2, "cmp", "immediate"       ],
0xca : [ 1, "dex", "implicit"        ],
0xcb : [ 2, "sbx", "immediate", und  ],
0xcc : [ 3, "cpy", "absolute"        ],
0xcd : [ 3, "cmp", "absolute"        ],
0xce : [ 3, "dec", "absolute"        ],
0xcf : [ 3, "dcp", "absolute", und   ],

0xd0 : [ 2, "bne", "relative", pcr   ],
0xd1 : [ 2, "cmp", "indirecty"       ],
0xd2 : [ 1, "hlt", "implicit", und   ],
0xd3 : [ 2, "dcp", "indirecty", und  ],
0xd4 : [ 2, "nop", "zeropagex", und  ],
0xd5 : [ 2, "cmp", "zeropagex"       ],
0xd6 : [ 2, "dec", "zeropagex"       ],
0xd7 : [ 2, "dcp", "zeropagex", und  ],
0xd8 : [ 1, "cld", "implicit"        ],
0xd9 : [ 3, "cmp", "absolutey"       ],
0xda : [ 1, "nop", "implicit", und   ],
0xdb : [ 3, "dcp", "absolutey", und  ],
0xdc : [ 3, "nop", "absolutex", und  ],
0xdd : [ 3, "cmp", "absolutex"       ],
0xde : [ 3, "dec", "absolutex"       ],
0xdf : [ 3, "dcp", "absolutex", und  ],

0xe0 : [ 2, "cpx", "immediate"       ],
0xe1 : [ 2, "sbc", "indirectx"       ],
0xe2 : [ 2, "nop", "immediate", und  ],
0xe3 : [ 2, "isc", "indirectx", und  ],
0xe4 : [ 2, "cpx", "zeropage"        ],
0xe5 : [ 2, "sbc", "zeropage"        ],
0xe6 : [ 2, "inc", "zeropage"        ],
0xe7 : [ 2, "isc", "zeropage", und   ],
0xe8 : [ 1, "inx", "implicit"        ],
0xe9 : [ 2, "sbc", "immediate"       ],
0xea : [ 1, "nop", "implicit"        ],
0xeb : [ 2, "sbc", "immediate", und  ],
0xec : [ 3, "cpx", "absolute"        ],
0xed : [ 3, "sbc", "absolute"        ],
0xee : [ 3, "inc", "absolute"        ],
0xef : [ 3, "isc", "absolute", und   ],

0xf0 : [ 2, "beq", "relative", pcr   ],
0xf1 : [ 2, "sbc", "indirecty"       ],
0xf2 : [ 1, "hlt", "implicit", und   ],
0xf3 : [ 2, "isc", "indirecty", und  ],
0xf4 : [ 2, "nop", "zeropagex", und  ],
0xf5 : [ 2, "sbc", "zeropagex"       ],
0xf6 : [ 2, "inc", "zeropagex"       ],
0xf7 : [ 2, "isc", "zeropagex", und  ],
0xf8 : [ 1, "sed", "implicit"        ],
0xf9 : [ 3, "sbc", "absolutey"       ],
0xfa : [ 1, "nop", "implicit", und   ],
0xfb : [ 3, "isc", "absolutey", und  ],
0xfc : [ 3, "nop", "absolutex", und  ],
0xfd : [ 3, "sbc", "absolutex"       ],
0xfe : [ 3, "inc", "absolutex"       ],
0xff : [ 3, "isc", "absolutex", und  ],

}

# End of processor specific code
##########################################################################



import sys

# Enthought library imports.
from envisage.ui.tasks.api import PreferencesPane, TaskFactory
from apptools.preferences.api import PreferencesHelper
from traits.api import Bool, Dict, Enum, List, Str, Unicode, Int, Font, Range
from traitsui.api import FontEditor, HGroup, VGroup, Item, Label, \
    View, RangeEditor

if sys.platform == "darwin":
    def_font = "10 point Monaco"
elif sys.platform == "win32":
    def_font = "10 point Lucida Console"
else:
    def_font = "10 point monospace"


class HexEditPreferences(PreferencesHelper):
    """ The preferences helper for the Framework application.
    """

    #### 'PreferencesHelper' interface ########################################

    # The path to the preference node that contains the preferences.
    preferences_path = 'omnivore.task.hex_edit'

    #### Preferences ##########################################################
    
    map_width_low = 1
    map_width_high = 256
    map_width = Range(low=map_width_low, high=map_width_high, value=16)
    
    bitmap_width_low = 1
    bitmap_width_high = 16
    bitmap_width = Range(low=bitmap_width_low, high=bitmap_width_high, value=1)

    # Font used for hex/disassembly
    text_font = Font(def_font)
    
    hex_grid_lower_case = Bool(True)
    
    assembly_lower_case = Bool(False)


class HexEditPreferencesPane(PreferencesPane):
    """ The preferences pane for the Framework application.
    """

    #### 'PreferencesPane' interface ##########################################

    # The factory to use for creating the preferences model object.
    model_factory = HexEditPreferences

    category = Str('Editors')

    #### 'FrameworkPreferencesPane' interface ################################

    # Note the quirk in the RangeEditor: specifying a custom editor is
    # supposed to take the defaults from the item name specified, but I
    # can't get it to work with only the "mode" parameter.  I have to specify
    # all the other params, and the low/high values have to be attributes
    # in HexEditPreferences, not the values in the trait itself.  See
    # traitsui/editors/range_editor.py
    view = View(
        VGroup(HGroup(Item('map_width', editor=RangeEditor(mode="spinner", is_float=False, low_name='map_width_low', high_name='map_width_high')),
                      Label('Default Character Map Width (in bytes)'),
                      show_labels = False),
               HGroup(Item('bitmap_width', editor=RangeEditor(mode="spinner", is_float=False, low_name='bitmap_width_low', high_name='bitmap_width_high')),
                      Label('Default Bitmap Width (in bytes)'),
                      show_labels = False),
               HGroup(Item('text_font'),
                      Label('Hex Display Font'),
                      show_labels = False),
               HGroup(Item('hex_grid_lower_case'),
                      Label('Use Lower Case for Hex Digits'),
                      show_labels = False),
               HGroup(Item('assembly_lower_case'),
                      Label('Use Lower Case for Assembler Mnemonics'),
                      show_labels = False),
               label='Hex Editor'),
        resizable=True)



import os
import itertools

import numpy as np

from pyface.tasks.topological_sort import before_after_sort
from omnivore.utils.sortutil import *

class SortItem(object):
    def __init__(self, id, before="", after=""):
        self.id = id
        self.before = before
        self.after = after

def check_sorted(items):
    order_lookup = {i.id: items.index(i) for i in items}
    print order_lookup
    wildcards = {}
    for item in items:
        for w in (item.before, item.after):
            if "*" in w and w not in wildcards:
                wildcards[w] = [0, 0]
    if wildcards:
        print "wildcards: %s" % str(wildcards)
        for w in wildcards.keys():
            start = w[:-1]
            pos = 0
            for item in items:
                if item.id.startswith(start):
                    wildcards[w][0] = pos
                    print "found first %s at %s: %s" % (w, pos, item.id)
                    break
                pos += 1
            pos = len(items)
            for item in reversed(items):
                pos -= 1
                if item.id.startswith(start):
                    wildcards[w][1] = pos
                    print "found last %s at %s: %s" % (w, pos, item.id)
                    break
        print "wildcards: %s" % str(wildcards)
    for item in items:
        pos = order_lookup[item.id]
        if item.before:
            if "*" in item.before:
                assert pos < wildcards[item.before][0]
            else:
                assert pos < order_lookup[item.before]
        if item.after:
            if "*" in item.after:
                assert pos > wildcards[item.after][1]
            else:
                assert pos > order_lookup[item.after]

class TestBeforeAfter(object):
    def setup(self):
        self.items = [
            SortItem('z'),
            SortItem('b', after='a'),
            SortItem('a', before='z'),
            ]

    def test_simple(self):
        items = before_after_sort(self.items)
        print ", ".join([i.id for i in items])
        check_sorted(items)

class TestWildcard(object):
    def setup(self):
        self.items = [
            SortItem('z/a'),
            SortItem('z/b'),
            SortItem('z/c'),
            SortItem('z/d'),
            SortItem('a/b', before='z/*'),
            SortItem('a/a', before='z/*'),
            SortItem('b/b', after='a/*'),
            SortItem('b/a'),
            SortItem('m/b', after='b/*'),
            SortItem('m/a'),
            ]

    def test_simple(self):
        items = before_after_wildcard_sort(self.items)
        print ", ".join([i.id for i in items])
        check_sorted(items)

class TestOverlappingRanges(object):
    def setup(self):
        self.items = [
            ([], []),
            ([(1, 2)], [(1, 2)]),
            ([(50, 90), (80, 182)], [(50, 182)]),
            ([(50, 90), (90, 182)], [(50, 182)]),
            ([(50, 90), (91, 182)], [(50, 90), (91, 182)]),
            ([(1, 100), (40, 82)], [(1, 100)]),
            ([(1, 100), (40, 82), (45, 55)], [(1, 100)]),
            ([(1, 100), (40, 82), (45, 55), (88, 200)], [(1, 200)]),
            ([(373, 440), (486, 537), (181, 182)], [(181, 182), (373, 440), (486, 537)]),
            ([(183, 182), (181, 182)], [(181, 183)]),
            ]

    def test_simple(self):
        for before, after in self.items:
            for p in itertools.permutations(before):
                processed = collapse_overlapping_ranges(p)
                print p, processed, after
                assert processed == after

class TestInvertRanges(object):
    def setup(self):
        self.items = [
            ([], 10, [(0, 10)]),
            ([(1, 2)], 10, [(0, 1), (2, 10)]),
            ([(50, 182)], 200, [(0, 50), (182, 200)]),
            ([(50, 90), (91, 182)], 200, [(0, 50), (90, 91), (182, 200)]),
            ([(181, 182), (373, 440), (486, 537)], 600, [(0, 181), (182, 373), (440, 486), (537, 600)]),
            ]

    def test_simple(self):
        for before, size, after in self.items:
            processed = invert_ranges(before, size)
            print size, before, processed, after
            assert processed == after
            processed = invert_ranges(after, size)
            assert processed == before
            print size, before, processed, after

class TestInvertRects(object):
    def setup(self):
        self.items = [
            ([[(5, 16), (8, 21)], ], 10, 50, [[(0, 0), (5, 16)], [(0, 16), (5, 21)], [(0, 21), (5, 50)], [(5, 0), (8, 16)], [(5, 21), (8, 50)], [(8, 0), (10, 16)], [(8, 16), (10, 21)], [(8, 21), (10, 50)]]),
            ]

    def test_simple(self):
        for before, rows, cols, after in self.items:
            processed = invert_rects(before, rows, cols)
            print rows, cols, before, processed, after
            assert processed == after

class TestRangeToIndex(object):
    def setup(self):
        self.items = [
            ([(50, 56)], (50, 51, 52, 53, 54, 55)),
            ([(50, 56), (60, 62)], (50, 51, 52, 53, 54, 55, 60, 61)),
            ([(50, 56), (60, 62), (70, 73)], (50, 51, 52, 53, 54, 55, 60, 61, 70, 71, 72)),
            ]

    def test_simple(self):
        for before, after in self.items:
            processed = ranges_to_indexes(before)
            after = np.asarray(after)
            print before, processed, after
            np.testing.assert_array_equal(processed, after)

if __name__ == "__main__":
    t = TestRangeToIndex()
    t.setup()
    t.test_simple()



import numpy as np

from omnivore.framework.errors import ProgressCancelError
from omnivore.utils.command import Batch, Command, UndoInfo
from omnivore.tasks.hex_edit.commands import ChangeByteCommand, SetValueCommand

import logging
progress_log = logging.getLogger("progress")


class CreateObjectCommand(SetValueCommand):
    short_name = "create_jumpman_obj"
    pretty_name = "Create Object"


class MoveObjectCommand(SetValueCommand):
    short_name = "move_jumpman_obj"
    pretty_name = "Move Object"


class FlipVerticalCommand(SetValueCommand):
    short_name = "vflip_jumpman_obj"
    pretty_name = "Flip Vertically"


class FlipHorizontalCommand(SetValueCommand):
    short_name = "hflip_jumpman_obj"
    pretty_name = "Flip Horizontally"



from omnivore.framework.panes import FrameworkPane

# Local imports.
from omnivore.tasks.hex_edit.segments import SegmentList
from omnivore.framework.undo_panel import UndoHistoryPanel

import logging
log = logging.getLogger(__name__)


class SegmentsPane(FrameworkPane):
    #### TaskPane interface ###################################################

    id = 'bitmap_edit.segments'
    name = 'Segments'
    
    def create_contents(self, parent):
        control = SegmentList(parent, self.task, size=(64,150))
        return control


class UndoPane(FrameworkPane):
    #### TaskPane interface ###################################################

    id = 'bitmap_edit.undo'
    name = 'Undo History'
    
    def create_contents(self, parent):
        control = UndoHistoryPanel(parent, self.task, size=(64,150))
        return control



import os
import sys

# Major package imports.
import numpy as np

# Enthought library imports.
from traits.api import HasTraits, Any, Bool, Int, Str, List, Dict, Event, Enum, DictStrStr

# Local imports.
import fonts
import colors
import disasm
import memory_map
import antic_renderers

import logging
log = logging.getLogger(__name__)


class Machine(HasTraits):
    """ Collection of classes that identify a machine: processor, display, etc.
    
    """
    
    # Traits
    
    name = Str

    mime_prefix = Str("application/octet-stream")
    
    disassembler = Any(transient=True)
    
    memory_map = Any(transient=True)
    
    antic_font_data = Any
    
    antic_font = Any(transient=True)

    blinking_antic_font = Any(transient=True)
    
    antic_color_registers = Any
    
    color_standard = Enum(0, 1)
    
    color_registers = Any(transient=True)
    
    color_registers_highlight = Any(transient=True)
    
    color_registers_data = Any(transient=True)
    
    color_registers_match = Any(transient=True)
    
    color_registers_comment = Any(transient=True)
    
    bitmap_renderer = Any(transient=True)
    
    font_renderer = Any(transient=True)
    
    font_mapping = Any(transient=True)
    
    page_renderer = Any(transient=True)
    
    assembler = Any(transient=True)
    
    # Trait events
    
    font_change_event = Event
    
    bitmap_shape_change_event = Event
    
    bitmap_color_change_event = Event
    
    disassembler_change_event = Event
    
    # Class attributes (not traits)
    
    font_list = None
    
    emulator_list = None
    
    assembler_list = None
    
    highlight_color = (100, 200, 230)
    
    unfocused_cursor_color = (128, 128, 128)
    
    background_color = (255, 255, 255)
    
    data_color = (224, 224, 224)
    
    match_background_color = (255, 255, 180)
    
    comment_background_color = (255, 180, 200)
    
    empty_color = None
    
    text_color = (0, 0, 0)
    
    diff_text_color = (255, 0, 0)
    
    text_font = None
        
    @classmethod
    def init_fonts(cls, editor):
        if cls.font_list is None:
            try:
                cls.font_list = editor.window.application.get_bson_data("font_list")
            except IOError:
                # file not found
                cls.font_list = []
            except ValueError:
                # bad JSON format
                cls.font_list = []
        prefs = editor.task.get_preferences()
        try:
            cls.text_font = prefs.text_font
        except AttributeError:
            pass
    
    @classmethod
    def remember_fonts(cls, application):
        application.save_bson_data("font_list", cls.font_list)
        
    @classmethod
    def init_emulators(cls, editor):
        if cls.emulator_list is None:
            cls.emulator_list = editor.window.application.get_json_data("emulator_list", [])
        default = editor.window.application.get_json_data("system_default_emulator", None)
        
        if default is None:
            default = cls.guess_system_default_emulator()
        if not cls.is_known_emulator(default):
            cls.emulator_list[0:0] = [default]
    
    @classmethod
    def remember_emulators(cls, application):
        e_list = []
        default = None
        for emu in cls.emulator_list:
            if 'system default' in emu:
                default = emu
            else:
                if 'system default' in emu:
                    # remove system default tags on any other emulator
                    del emu['system default']
                e_list.append(emu)
        if e_list:
            application.save_json_data("emulator_list", e_list)
        if default:
            application.save_json_data("system_default_emulator", default)
        
    @classmethod
    def init_assemblers(cls, editor):
        if cls.assembler_list is None:
            cls.assembler_list = editor.window.application.get_json_data("assembler_list", [])
        
        if not cls.assembler_list:
            cls.assembler_list = cls.guess_default_assemblers()
    
    @classmethod
    def remember_assemblers(cls, application):
        if cls.assembler_list:
            application.save_json_data("assembler_list", cls.assembler_list)
    
    @classmethod
    def init_colors(cls, editor):
        if cls.empty_color is None:
            attr = editor.control.GetDefaultAttributes()
            cls.empty_color = attr.colBg.Get(False)
    
    @classmethod
    def one_time_init(cls, editor):
        cls.init_fonts(editor)
        cls.init_colors(editor)
        cls.init_emulators(editor)
        cls.init_assemblers(editor)
    
    @classmethod
    def set_text_font(cls, editor, font, color):
        cls.text_color = color
        cls.text_font = font
        prefs = editor.task.get_preferences()
        prefs.text_font = font

    @classmethod
    def find_machine_by_mime(cls, mime):
        for m in predefined['machine']:
            if mime.startswith(m.mime_prefix):
                return m
    
    # Trait initializers
    
    def _name_default(self):
        return "Generic 6502"
    
    def _assembler_default(self):
        for asm in self.assembler_list:
            if 'system default' in asm:
                return asm
        return self.assembler_list[0]
    
    def _disassembler_default(self):
        return predefined['disassembler'][0]

    def _memory_map_default(self):
        return predefined['memory_map'][0]
    
    def _antic_font_default(self):
        return self.get_antic_font()
    
    def _blinking_antic_font_default(self):
        return self.get_antic_font(True)
    
    def _antic_font_data_default(self):
        return fonts.A8DefaultFont
    
    def _antic_color_registers_default(self):
        return list(colors.powerup_colors())
    
    def _color_standard_default(self):
        return 0  # NTSC
    
    def _color_registers_default(self):
        return self.get_color_registers()
    
    def _color_registers_highlight_default(self):
        return self.get_blended_color_registers(self.color_registers, self.highlight_color)
    
    def _color_registers_match_default(self):
        return self.get_blended_color_registers(self.color_registers, self.match_background_color)
    
    def _color_registers_comment_default(self):
        return self.get_blended_color_registers(self.color_registers, self.comment_background_color)
    
    def _color_registers_data_default(self):
        return self.get_dimmed_color_registers(self.color_registers, self.background_color, self.data_color)

    def _bitmap_renderer_default(self):
        return predefined['bitmap_renderer'][0]
    
    def _font_renderer_default(self):
        return predefined['font_renderer'][0]
    
    def _font_mapping_default(self):
        return predefined['font_mapping'][0]
    
    def _page_renderer_default(self):
        return predefined['page_renderer'][0]
    
    def __getstate__(self):
        state = super(Machine, self).__getstate__()
        
        for name in self.all_trait_names():
            t = self.trait(name)
            if t.transient:
                if name in predefined:
                    value = getattr(self, name)
                    try:
                        state[name] = predefined[name].index(value)
                    except ValueError:
                        log.warning("No matching index for predefined[%s]; will use default upon load instead of %s" % (name, value))
            elif name == "memory_map":
                # convert into list of tuples so json won't mangle the integer
                # keys into strings
                value = getattr(self, name)
                state[name] = value.items()
        
        print state.keys()
        return state
    
    def state_to_traits(self, state):
        for name in self.all_trait_names():
            if name not in state:
                # skip missing trait definitions
                continue
            t = self.trait(name)
            if t.transient:
                if name in predefined:
                    index = state[name]
                    setattr(self, name, predefined[name][index])
            elif name == "memory_map":
                # convert into list of tuples so json won't mangle the integer
                # keys into strings
                value = state[name]
                setattr(self, name, {k:v for k,v in value})
            else:
                try:
                    setattr(self, name, state[name])
                except KeyError:
                    pass

    def __setstate__(self, state):
        self.state_to_traits(state)
        # set up trait notifications
        self._init_trait_listeners()
    # 
    
    def __eq__(self, other):
        return self.disassembler == other.disassembler and self.memory_map == other.memory_map

    def clone_machine(self):
        m = self.clone_traits()
        m.update_colors(m.antic_color_registers)
        m.set_font()
        return m

    def serialize_extra_to_dict(self, mdict):
        """Save extra metadata to a dict so that it can be serialized
        """
        s = self.__getstate__()
        mdict.update(s)

    def restore_extra_from_dict(self, e):
        self.state_to_traits(e)
        self.update_colors(self.antic_color_registers)
    #
    
    def update_colors(self, c):
        baseline = list(colors.powerup_colors())
        # need to operate on a copy of the colors to make sure we're not
        # changing some global value. Also force as python int so we're not
        # mixing numpy and python values.
        if len(c) == 5:
            baseline[4:9] = [int(i) for i in c]
        else:
            baseline[0:len(c)] = [int(i) for i in c]
        self.antic_color_registers = baseline
        self.color_registers = self.get_color_registers()
        self.color_registers_highlight = self.get_blended_color_registers(self.color_registers, self.highlight_color)
        self.color_registers_match = self.get_blended_color_registers(self.color_registers, self.match_background_color)
        self.color_registers_comment = self.get_blended_color_registers(self.color_registers, self.comment_background_color)
        self.set_font()
        self.bitmap_color_change_event = True
    
    def get_color_registers(self, antic_color_registers=None):
        color_converter = self.get_color_converter()
        registers = []
        if antic_color_registers is None:
            antic_color_registers = self.antic_color_registers
        for c in antic_color_registers:
            registers.append(color_converter(c))
        
        # make sure there are 16 registers for 4bpp modes
        i = len(registers)
        for i in range(len(registers), 16):
            registers.append((i*16, i*16, i*16))

        # Extend to 32 for dimmed copies of the 16 colors
        dim = []
        for r in registers:
            dim.append((r[0]/4 + 64, r[1]/4 + 64, r[2]/4 + 64))
        registers.extend(dim)
        return registers
    
    def get_blended_color_registers(self, colors, blend_color):
        registers = []
        base_blend = [(r * 7)/8 for r in blend_color]
        for c in colors:
            r = [c[i]/8 + base_blend[i] for i in range(3)]
            registers.append(r)
        return registers
    
    def get_dimmed_color_registers(self, colors, background_color, dimmed_color):
        registers = []
        dimmed_difference = [b - d for b, d in zip(background_color, dimmed_color)]
        for c in colors:
            r = [max(0, c[i]- dimmed_difference[i]) for i in range(3)]
            registers.append(r)
        return registers
    
    def get_color_converter(self):
        if self.color_standard == 0:
            return colors.gtia_ntsc_to_rgb
        return colors.gtia_pal_to_rgb
    
    def set_color_standard(self, std):
        self.color_standard = std
        self.update_colors(self.antic_color_registers)
    
    def set_bitmap_renderer(self, renderer):
        self.bitmap_renderer = renderer
        self.bitmap_shape_change_event = True
    
    def set_disassembler(self, disassembler):
        self.disassembler = disassembler
        self.disassembler_change_event = True
    
    def set_memory_map(self, memory_map):
        self.memory_map = memory_map
        self.disassembler_change_event = True
    
    def set_font(self, font=None, font_renderer=None):
        if font is None:
            font = self.antic_font_data
        if font_renderer is not None:
            if isinstance(font_renderer, str):
                font_renderer = self.get_font_renderer_from_font_name(font_renderer)
            self.font_renderer = font_renderer
        self.antic_font_data = font
        self.antic_font = self.get_antic_font()
        if self.antic_font.use_blinking:
            self.blinking_antic_font = self.get_antic_font(True)
        else:
            self.blinking_antic_font = None
        self.set_font_mapping()

    def get_blinking_font(self, index):
        if self.antic_font.use_blinking and index == 1 and self.blinking_antic_font is not None:
            return self.blinking_antic_font
        else:
            return self.antic_font
    
    def get_font_renderer_from_font_name(self, font_name):
        for r in predefined['font_renderer']:
            if r.name.startswith(font_name):
                return r
        return predefined['font_renderer'][0]
    
    def set_font_mapping(self, font_mapping=None):
        if font_mapping is None:
            font_mapping = self.font_mapping
        self.font_mapping = font_mapping
        self.font_change_event = True
    
    def get_antic_font(self, reverse=False):
        return fonts.AnticFont(self, self.antic_font_data, self.font_renderer, self.antic_color_registers[4:9], reverse)
    
    def load_font(self, task, filename):
        try:
            fh = open(filename, 'rb')
            data = fh.read() + "\0"*1024
            data = data[0:1024]
            font = {
                'name': os.path.basename(filename),
                'data': data,
                'char_w': 8,
                'char_h': 8,
                }
            self.set_font(font)
            self.font_list.append(font)
            self.remember_fonts(task.window.application)
            task.machine_menu_changed = self
        except:
            raise
    
    def add_emulator(self, task, emu):
        self.emulator_list.append(emu)
        self.remember_emulators(task.window.application)
        task.machine_menu_changed = self
    
    @classmethod
    def is_known_emulator(cls, emu):
        for e in cls.emulator_list:
            if e == emu:
                return True
        return False
    
    @classmethod
    def guess_system_default_emulator(cls):
        if sys.platform == "win32":
            exe = "Altirra.exe"
        elif sys.platform == "darwin":
            exe = "Atari800MacX"
        else:
            exe = "atari800"
        emu = {'exe': exe,
               'args': "",
               'name': "system default: %s" % exe,
               'system default': True,
               }
        return emu
    
    @classmethod
    def get_system_default_emulator(cls, task):
        try:
            default = cls.emulator_list[0]
        except IndexError:
            # somehow, all the elements have been removed!
            default = cls.guess_system_default_emulator()
            cls.remember_emulators(task.window.application)
            task.machine_menu_changed = cls
        return default
    
    @classmethod
    def set_system_default_emulator(cls, task, emu):
        emu = dict(emu)  # copy to make sure we're not referencing an item in the existing emulator_list
        emu['system default'] = True
        emu['name'] = "system default: %s" % emu['name']
        default = cls.emulator_list[0]
        if 'system default' not in default:
            cls.emulator_list[0:0] = [emu]
        else:
            cls.emulator_list[0] = emu
        cls.remember_emulators(task.window.application)
        task.machine_menu_changed = cls
    
    @classmethod
    def get_user_defined_emulator_list(cls):
        """Return list of user defined emulators (i.e. not including the system
        default emulator
        """
        emus = []
        for e in cls.emulator_list:
            if 'system default' not in e:
                emus.append(e)
        return emus
    
    @classmethod
    def set_user_defined_emulator_list(cls, task, emus):
        default = None
        for e in cls.emulator_list:
            if 'system default' in e:
                default = e
        emus[0:0] = [default]
        cls.emulator_list = emus
        cls.remember_emulators(task.window.application)
        task.machine_menu_changed = cls
    
    def set_assembler(self, assembler):
        self.assembler = assembler
        self.disassembler_change_event = True
    
    def add_assembler(self, task, d):
        self.assembler_list.append(d)
        self.remember_assemblers(task.window.application)
        task.machine_menu_changed = self
    
    @classmethod
    def guess_default_assemblers(cls):
        asm_list = [
            {'comment char': ';',
             'origin': '.org',
             'data byte': '.byte',
             'name': "cc65",
             },
            {'comment char': ';',
             'origin': '*=',
             'data byte': '.byte',
             'name': "MAC/65",
             },
            ]
        return [dict(asm) for asm in asm_list]  # force a copy
    
    @classmethod
    def set_system_default_assembler(cls, task, asm):
        if 'system default' in asm:
            del asm['system default']
        for e in cls.assembler_list:
            if 'system default' in e:
                del e['system default']
        found = False
        for e in cls.assembler_list:
            if e == asm:
                asm['system default'] = True
                found = True
                break
        if not found:
            cls.assembler_list[0:0] = [asm]
        cls.remember_assemblers(task.window.application)
        task.machine_menu_changed = cls
    
    @classmethod
    def get_default_assembler(cls):
        found = None
        for e in cls.assembler_list:
            if 'system default' in e:
                found = e
        if not found:
            if cls.assembler_list:
                found = cls.assembler_list[0]
        return found

    @classmethod
    def set_assembler_list(cls, task, asms):
        default = None
        for e in asms:
            if 'system default' in e:
                if default:
                    del e['system default']
                else:
                    default = e
        cls.assembler_list = asms
        cls.remember_assemblers(task.window.application)
        task.machine_menu_changed = cls
    
    def verify_current_assembler(self):
        found = None
        for asm in self.assembler_list:
            if self.assembler == asm:
                found = asm
                break
        if not found:
            asm = self.get_default_assembler()
        self.set_assembler(asm)
    
    # Utility methods
    
    def get_disassembler(self, hex_lower, mnemonic_lower):
        return self.disassembler(self.assembler, self.memory_map, hex_lower, mnemonic_lower)

    def get_nop(self):
        return self.disassembler.get_nop()


Generic6502 = Machine(name="Generic 6502", disassembler=disasm.Basic6502Disassembler)

Atari800 = Machine(name="Atari 800", mime_prefix="application/vnd.atari8bit", disassembler=disasm.Basic6502Disassembler, memory_map=memory_map.Atari800MemoryMap)

Atari800Undoc = Machine(name="Atari 800 (show undocumented opcodes)", mime_prefix="application/vnd.atari8bit", disassembler=disasm.Undocumented6502Disassembler, memory_map=memory_map.Atari800MemoryMap)

Atari800Flagged = Machine(name="Atari 800 (highlight undocumented opcodes)", mime_prefix="application/vnd.atari8bit", disassembler=disasm.Flagged6502Disassembler, memory_map=memory_map.Atari800MemoryMap)

Atari5200 = Machine(name="Atari 5200", mime_prefix="application/vnd.atari8bit", disassembler=disasm.Basic6502Disassembler, memory_map=memory_map.Atari5200MemoryMap)



predefined = {
    "machine": [
        Generic6502,
        Atari800,
        Atari800Undoc,
        Atari800Flagged,
        Atari5200,
        ],
    "memory_map": [
        memory_map.EmptyMemoryMap,
        memory_map.Atari800MemoryMap,
        memory_map.Atari5200MemoryMap,
        memory_map.Apple2MemoryMap,
        ],
    "disassembler": [
        disasm.Basic6502Disassembler,
        disasm.Undocumented6502Disassembler,
        disasm.Flagged6502Disassembler,
        disasm.Basic65C02Disassembler,
        disasm.Basic65816Disassembler,
        disasm.Basic6800Disassembler,
        disasm.Basic6809Disassembler,
        disasm.Basic6811Disassembler,
        disasm.Basic8051Disassembler,
        disasm.Basic8080Disassembler,
        disasm.BasicZ80Disassembler,
        ],
    "bitmap_renderer": [
        antic_renderers.OneBitPerPixelB(),
        antic_renderers.OneBitPerPixelW(),
        antic_renderers.OneBitPerPixelPM1(),
        antic_renderers.OneBitPerPixelPM2(),
        antic_renderers.OneBitPerPixelPM4(),
        antic_renderers.OneBitPerPixelApple2(),
        antic_renderers.ModeB(),
        antic_renderers.ModeC(),
        antic_renderers.ModeD(),
        antic_renderers.ModeE(),
        antic_renderers.GTIA9(),
        antic_renderers.GTIA10(),
        antic_renderers.GTIA11(),
        antic_renderers.TwoBitsPerPixel(),
        antic_renderers.FourBitsPerPixel(),
        antic_renderers.TwoBitPlanesLE(),
        antic_renderers.TwoBitPlanesLineLE(),
        antic_renderers.TwoBitPlanesBE(),
        antic_renderers.TwoBitPlanesLineBE(),
        antic_renderers.ThreeBitPlanesLE(),
        antic_renderers.ThreeBitPlanesLineLE(),
        antic_renderers.ThreeBitPlanesBE(),
        antic_renderers.ThreeBitPlanesLineBE(),
        antic_renderers.FourBitPlanesLE(),
        antic_renderers.FourBitPlanesLineLE(),
        antic_renderers.FourBitPlanesBE(),
        antic_renderers.FourBitPlanesLineBE(),
        ],
    "font_renderer": [
        antic_renderers.Mode2(),
        antic_renderers.Mode4(),
        antic_renderers.Mode5(),
        antic_renderers.Mode6Upper(),
        antic_renderers.Mode6Lower(),
        antic_renderers.Mode7Upper(),
        antic_renderers.Mode7Lower(),
        antic_renderers.Apple2TextMode(),
        ],
    "font_mapping": [
        antic_renderers.ATASCIIFontMapping(),
        antic_renderers.AnticFontMapping(),
        ],
    "page_renderer": [
        antic_renderers.BytePerPixelMemoryMap(),
        ],
    }


Apple2 = Machine(name="Apple ][", mime_prefix="application/vnd.apple2", disassembler=disasm.Basic65C02Disassembler, antic_font_data=fonts.A2DefaultFont, font_renderer=predefined['font_renderer'][7], font_mapping=predefined['font_mapping'][1], antic_color_registers=[4, 30, 68, 213, 15, 202, 148, 70, 0], memory_map=memory_map.Apple2MemoryMap)
predefined['machine'].append(Apple2)



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought pyface package component>
#------------------------------------------------------------------------------
""" A dialog that allows the user to chose a single item from a list. """
from __future__ import absolute_import

# Major package imports.
import wx

# Enthought library imports.
from traits.api import List, Str, Any

# Local imports.
from .dialog import Dialog


class SingleChoiceDialog(Dialog):
    """ A dialog that allows the user to chose a single item from a list.

    choices is the list of things to choose from.   If name_attribute is
    set then we assume it is a list of objects getattr(obj, name_attribute)
    gives us the string to put in the dialog.  Otherwise we just call str()
    on the choices to get the strings.
    """

    choices = List(Any)
    choice = Any
    name_attribute = Str
    caption = Str

    ###########################################################################
    # 'Window' interface.
    ###########################################################################

    def close(self):
        """ Closes the window. """

        # Get the chosen object.
        if self.control is not None:
            self.choice = self.choices[self.control.GetSelection()]

        # Let the window close as normal.
        super(SingleChoiceDialog, self).close()

        return

    ###########################################################################
    # Protected 'Window' interface.
    ###########################################################################

    def _create_control(self, parent):
        """ Create the toolkit-specific control that represents the window. """

        dialog = wx.SingleChoiceDialog(
            parent,
            self.title,
            self.caption,
            self._get_string_choices(),
            self.STYLE
        )

        return dialog

    def _create_contents(self, parent):
        """ Creates the window contents. """

        # In this case, wx does it all for us in 'wx.SingleChoiceDialog'
        pass

    ###########################################################################
    # 'Private' interface.
    ###########################################################################

    def _get_string_choices(self):
        """ Returns the list of strings to display in the dialog. """

        if len(self.name_attribute) > 0:
            # We asssume choices is a list of objects with this attribute
            choices = [
                getattr(obj, self.name_attribute) for obj in self.choices
            ]

        else:
            # We just convert to strings
            choices = [str(obj) for obj in self.choices]

        return choices

#### EOF ######################################################################



#------------------------------------------------------------------------------
# Copyright (c) 2008, Riverbank Computing Limited
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD license.
# However, when used with the GPL version of PyQt the additional terms described in the PyQt GPL exception also apply

#
# Author: Riverbank Computing Limited
#------------------------------------------------------------------------------

""" Defines the various editors and the editor factory for single-selection
    enumerations, for the PyQt user interface toolkit.
"""

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

from pyface.qt import QtCore, QtGui

from traits.api \
    import Bool, Property

# FIXME: ToolkitEditorFactory is a proxy class defined here just for backward
# compatibility. The class has been moved to the
# traitsui.editors.enum_editor file.
from traitsui.editors.enum_editor \
    import ToolkitEditorFactory

from editor \
    import Editor

from constants \
    import OKColor, ErrorColor

from traitsui.helper \
    import enum_values_changed


# default formatting function (would import from string, but not in Python 3)
capitalize = lambda s: s.capitalize()


#-------------------------------------------------------------------------------
#  'BaseEditor' class:
#-------------------------------------------------------------------------------

class BaseEditor ( Editor ):
    """ Base class for enumeration editors.
    """

    #---------------------------------------------------------------------------
    #  Trait definitions:
    #---------------------------------------------------------------------------

    # Current set of enumeration names:
    names = Property

    # Current mapping from names to values:
    mapping = Property

    # Current inverse mapping from values to names:
    inverse_mapping = Property

    #---------------------------------------------------------------------------
    #  Finishes initializing the editor by creating the underlying toolkit
    #  widget:
    #---------------------------------------------------------------------------

    def init ( self, parent ):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        factory = self.factory
        if factory.name != '':
            self._object, self._name, self._value = \
                self.parse_extended_name( factory.name )
            self.values_changed()
            self._object.on_trait_change( self._values_changed,
                                          ' ' + self._name, dispatch = 'ui' )
        else:
            factory.on_trait_change( self.rebuild_editor, 'values_modified',
                                     dispatch = 'ui' )

    #---------------------------------------------------------------------------
    #  Gets the current set of enumeration names:
    #---------------------------------------------------------------------------

    def _get_names ( self ):
        """ Gets the current set of enumeration names.
        """
        if self._object is None:
            return self.factory._names

        return self._names

    #---------------------------------------------------------------------------
    #  Gets the current mapping:
    #---------------------------------------------------------------------------

    def _get_mapping ( self ):
        """ Gets the current mapping.
        """
        if self._object is None:
            return self.factory._mapping

        return self._mapping

    #---------------------------------------------------------------------------
    #  Gets the current inverse mapping:
    #---------------------------------------------------------------------------

    def _get_inverse_mapping ( self ):
        """ Gets the current inverse mapping.
        """
        if self._object is None:
            return self.factory._inverse_mapping

        return self._inverse_mapping

    #---------------------------------------------------------------------------
    #  Rebuilds the contents of the editor whenever the original factory
    #  object's 'values' trait changes:
    #---------------------------------------------------------------------------

    def rebuild_editor ( self ):
        """ Rebuilds the contents of the editor whenever the original factory
            object's **values** trait changes.
        """
        raise NotImplementedError

    #---------------------------------------------------------------------------
    #  Recomputes the cached data based on the underlying enumeration model:
    #---------------------------------------------------------------------------

    def values_changed ( self ):
        """ Recomputes the cached data based on the underlying enumeration model.
        """
        self._names, self._mapping, self._inverse_mapping = \
            enum_values_changed( self._value(), self.string_value )

    #---------------------------------------------------------------------------
    #  Handles the underlying object model's enumeration set being changed:
    #---------------------------------------------------------------------------

    def _values_changed ( self ):
        """ Handles the underlying object model's enumeration set being changed.
        """
        self.values_changed()
        self.rebuild_editor()

    #---------------------------------------------------------------------------
    #  Disposes of the contents of an editor:
    #---------------------------------------------------------------------------

    def dispose ( self ):
        """ Disposes of the contents of an editor.
        """
        if self._object is not None:
            self._object.on_trait_change( self._values_changed,
                                          ' ' + self._name, remove = True )
        else:
            self.factory.on_trait_change( self.rebuild_editor,
                                          'values_modified', remove = True )

        super( BaseEditor, self ).dispose()

#-------------------------------------------------------------------------------
#  'SimpleEditor' class:
#-------------------------------------------------------------------------------

class SimpleEditor ( BaseEditor ):
    """ Simple style of enumeration editor, which displays a combo box.
    """

    #---------------------------------------------------------------------------
    #  Finishes initializing the editor by creating the underlying toolkit
    #  widget:
    #---------------------------------------------------------------------------

    def init ( self, parent ):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        super( SimpleEditor, self ).init( parent )

        self.control = control = self.create_combo_box()
        control.addItems(self.names)

        QtCore.QObject.connect(control,
                               QtCore.SIGNAL('currentIndexChanged(QString)'),
                               self.update_object)

        if self.factory.evaluate is not None:
            control.setEditable(True)
            if self.factory.auto_set:
                QtCore.QObject.connect(control,
                                       QtCore.SIGNAL('editTextChanged(QString)'),
                                       self.update_text_object)
            else:
                QtCore.QObject.connect(control.lineEdit(),
                                   QtCore.SIGNAL('editingFinished()'),
                                   self.update_autoset_text_object)
            control.setInsertPolicy(QtGui.QComboBox.NoInsert)

        self._no_enum_update = 0
        self.set_tooltip()

    #---------------------------------------------------------------------------
    #  Returns the QComboBox used for the editor control:
    #---------------------------------------------------------------------------

    def create_combo_box(self):
        """ Returns the QComboBox used for the editor control.
        """
        control = QtGui.QComboBox()
        control.setSizeAdjustPolicy(QtGui.QComboBox.AdjustToContents)
        control.setSizePolicy(QtGui.QSizePolicy.Maximum,
                              QtGui.QSizePolicy.Fixed)
        return control

    #---------------------------------------------------------------------------
    #  Adjust size policy to behave properly in group
    #---------------------------------------------------------------------------

    def set_size_policy(self, direction, resizable, springy, stretch) :
        super(SimpleEditor, self).set_size_policy(direction, resizable, springy, stretch)

        if ((direction == QtGui.QBoxLayout.LeftToRight and springy) or
            (direction != QtGui.QBoxLayout.LeftToRight and resizable)) :
            self.control.setSizeAdjustPolicy(QtGui.QComboBox.AdjustToContentsOnFirstShow)

    #---------------------------------------------------------------------------
    #  Handles the user selecting a new value from the combo box:
    #---------------------------------------------------------------------------

    def update_object (self, text):
        """ Handles the user selecting a new value from the combo box.
        """
        if self._no_enum_update == 0:
            self._no_enum_update += 1
            try:
                self.value = self.mapping[unicode(text)]
            except:
                from traitsui.api import raise_to_debug
                raise_to_debug()
            self._no_enum_update -= 1

    #---------------------------------------------------------------------------
    #  Handles the user typing text into the combo box text entry field:
    #---------------------------------------------------------------------------

    def update_text_object(self, text):
        """ Handles the user typing text into the combo box text entry field.
        """
        if self._no_enum_update == 0:
            value = unicode(text)
            try:
                value = self.mapping[value]
            except:
                try:
                    value = self.factory.evaluate(value)
                except Exception as excp:
                    self.error( excp )
                    return

            self._no_enum_update += 1
            self.value = value
            self._set_background(OKColor)
            self._no_enum_update -= 1

    def update_autoset_text_object(self):
        # Don't get the final text with the editingFinished signal
        if self.control is not None:
            text = self.control.lineEdit().text()
            return self.update_text_object(text)

    #---------------------------------------------------------------------------
    #  Updates the editor when the object trait changes external to the editor:
    #---------------------------------------------------------------------------

    def update_editor ( self ):
        """ Updates the editor when the object trait changes externally to the
            editor.
        """
        if self._no_enum_update == 0:
            self._no_enum_update += 1
            if self.factory.evaluate is None:
                try:
                    index = self.names.index(self.inverse_mapping[self.value])
                    self.control.setCurrentIndex(index)
                except:
                    self.control.setCurrentIndex(-1)
            else:
                try:
                    self.control.setEditText(self.str_value)
                except:
                    self.control.setEditText('')
            self._no_enum_update -= 1

    #---------------------------------------------------------------------------
    #  Handles an error that occurs while setting the object's trait value:
    #---------------------------------------------------------------------------

    def error ( self, excp ):
        """ Handles an error that occurs while setting the object's trait value.
        """
        self._set_background(ErrorColor)

    #---------------------------------------------------------------------------
    #  Sets the background color of the QLineEdit of the QComboBox.
    #---------------------------------------------------------------------------

    def _set_background(self, col):
        le = self.control.lineEdit()
        pal = QtGui.QPalette(le.palette())
        pal.setColor(QtGui.QPalette.Base, col)
        le.setPalette(pal)

    #---------------------------------------------------------------------------
    #  Rebuilds the contents of the editor whenever the original factory
    #  object's 'values' trait changes:
    #---------------------------------------------------------------------------

    def rebuild_editor ( self ):
        """ Rebuilds the contents of the editor whenever the original factory
            object's **values** trait changes.
        """
        self.control.blockSignals(True)
        self.control.clear()
        self.control.addItems(self.names)
        self.control.blockSignals(False)

        self.update_editor()

#-------------------------------------------------------------------------------
#  'RadioEditor' class:
#-------------------------------------------------------------------------------

class RadioEditor ( BaseEditor ):
    """ Enumeration editor, used for the "custom" style, that displays radio
        buttons.
    """

    # Is the button layout row-major or column-major?
    row_major = Bool( False )

    #---------------------------------------------------------------------------
    #  Finishes initializing the editor by creating the underlying toolkit
    #  widget:
    #---------------------------------------------------------------------------

    def init ( self, parent ):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        super( RadioEditor, self ).init( parent )

        self.control = QtGui.QWidget()
        layout = QtGui.QGridLayout(self.control)
        layout.setContentsMargins(0, 0, 0, 0)

        self._mapper = QtCore.QSignalMapper()
        QtCore.QObject.connect(self._mapper, QtCore.SIGNAL('mapped(int)'),
                               self.update_object)

        self.rebuild_editor()

    #---------------------------------------------------------------------------
    #  Handles the user clicking one of the 'custom' radio buttons:
    #---------------------------------------------------------------------------

    def update_object ( self, index ):
        """ Handles the user clicking one of the custom radio buttons.
        """
        try:
            self.value = self.mapping[self.names[index]]
        except:
            pass

    #---------------------------------------------------------------------------
    #  Updates the editor when the object trait changes external to the editor:
    #---------------------------------------------------------------------------

    def update_editor ( self ):
        """ Updates the editor when the object trait changes externally to the
            editor.
        """
        layout = self.control.layout()
        value = self.value
        for i in range(layout.count()):
            rb = layout.itemAt(i).widget()
            rb.setChecked(rb.value == value)

    #---------------------------------------------------------------------------
    #  Rebuilds the contents of the editor whenever the original factory
    #  object's 'values' trait changes:
    #---------------------------------------------------------------------------

    def rebuild_editor ( self ):
        """ Rebuilds the contents of the editor whenever the original factory
            object's **values** trait changes.
        """
        # Clear any existing content:
        self.clear_layout()

        # Get the current trait value:
        cur_name = self.str_value

        # Create a sizer to manage the radio buttons:
        names   = self.names
        mapping = self.mapping
        n       = len( names )
        cols    = self.factory.cols
        rows    = (n + cols - 1) // cols
        if self.row_major:
            incr = [ 1 ] * cols
        else:
            incr = [ n / cols ] * cols
            rem  = n % cols
            for i in range( cols ):
                incr[i] += (rem > i)
            incr[-1] = -(reduce( lambda x, y: x + y, incr[:-1], 0 ) - 1)

        # Add the set of all possible choices:
        layout = self.control.layout()
        index = 0
        for i in range( rows ):
            for j in range( cols ):
                if n > 0:
                    name = names[index]
                    rb = self.create_button(name)
                    rb.value = mapping[name]

                    rb.setChecked(name == cur_name)

                    QtCore.QObject.connect(rb, QtCore.SIGNAL('clicked()'),
                                           self._mapper, QtCore.SLOT('map()'))
                    self._mapper.setMapping(rb, index)

                    self.set_tooltip(rb)
                    layout.addWidget(rb, i, j)

                    index += incr[j]
                    n -= 1

    #---------------------------------------------------------------------------
    #  Returns the QAbstractButton used for the radio button:
    #---------------------------------------------------------------------------

    def create_button(self, name):
        """ Returns the QAbstractButton used for the radio button.
        """
        label = self.string_value(name, capitalize)
        return QtGui.QRadioButton(label)

#-------------------------------------------------------------------------------
#  'ListEditor' class:
#-------------------------------------------------------------------------------

class ListEditor ( BaseEditor ):
    """ Enumeration editor, used for the "custom" style, that displays a list
        box.
    """
    #---------------------------------------------------------------------------
    #  Finishes initializing the editor by creating the underlying toolkit
    #  widget:
    #---------------------------------------------------------------------------

    def init ( self, parent ):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        super( ListEditor, self ).init( parent )

        self.control = QtGui.QListWidget()
        QtCore.QObject.connect(self.control,
                QtCore.SIGNAL('currentTextChanged(QString)'),
                self.update_object)

        self.rebuild_editor()
        self.set_tooltip()

    #---------------------------------------------------------------------------
    #  Handles the user selecting a list box item:
    #---------------------------------------------------------------------------

    def update_object(self, text):
        """ Handles the user selecting a list box item.
        """
        value = unicode(text)
        try:
            value = self.mapping[ value ]
        except:
            try:
                value = self.factory.evaluate( value )
            except:
                pass
        try:
            self.value = value
        except:
            pass

    #---------------------------------------------------------------------------
    #  Updates the editor when the object trait changes external to the editor:
    #---------------------------------------------------------------------------

    def update_editor ( self ):
        """ Updates the editor when the object trait changes externally to the
            editor.
        """
        control = self.control
        try:
            value = self.inverse_mapping[self.value]

            for row in range(control.count()):
                itm = control.item(row)

                if itm.text() == value:
                    control.setCurrentItem(itm)
                    control.scrollToItem(itm)
                    break
        except:
            pass

    #---------------------------------------------------------------------------
    #  Rebuilds the contents of the editor whenever the original factory
    #  object's 'values' trait changes:
    #---------------------------------------------------------------------------

    def rebuild_editor ( self ):
        """ Rebuilds the contents of the editor whenever the original factory
            object's **values** trait changes.
        """

        self.control.blockSignals(True)
        self.control.clear()
        for name in self.names:
            self.control.addItem(name)
        self.control.blockSignals(False)

        self.update_editor()



#------------------------------------------------------------------------------
#
#  Copyright (c) 2009, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#------------------------------------------------------------------------------

""" Defines the HTML "editor" for the QT4 user interface toolkit.
    HTML editors interpret and display HTML-formatted text, but do not
    modify it.
"""

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

import webbrowser

from pyface.qt import QtCore, QtGui, QtWebKit

from traits.api import Str

from editor import Editor

#-------------------------------------------------------------------------------
#  'SimpleEditor' class:
#-------------------------------------------------------------------------------

class SimpleEditor ( Editor ):
    """ Simple style editor for HTML.
    """

    #---------------------------------------------------------------------------
    #  Trait definitions:
    #---------------------------------------------------------------------------

    # Is the HTML editor scrollable? This values override the default.
    scrollable = True

    # External objects referenced in the HTML are relative to this URL
    base_url = Str

    #---------------------------------------------------------------------------
    #  Finishes initializing the editor by creating the underlying toolkit
    #  widget:
    #---------------------------------------------------------------------------

    def init ( self, parent ):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        self.control = QtWebKit.QWebView()
        self.control.setSizePolicy( QtGui.QSizePolicy.Expanding,
                                    QtGui.QSizePolicy.Expanding )

        if self.factory.open_externally:
            page = self.control.page()
            page.setLinkDelegationPolicy( QtWebKit.QWebPage.DelegateAllLinks )
            signal = QtCore.SIGNAL( 'linkClicked(QUrl)' )
            QtCore.QObject.connect( page, signal, self._link_clicked )

        self.base_url = self.factory.base_url
        self.sync_value( self.factory.base_url_name, 'base_url', 'from' )

    #---------------------------------------------------------------------------
    #  Updates the editor when the object trait changes external to the editor:
    #---------------------------------------------------------------------------

    def update_editor ( self ):
        """ Updates the editor when the object trait changes external to the
            editor.
        """
        text = self.str_value
        if self.factory.format_text:
            text = self.factory.parse_text( text )
        if self.base_url:
            url = self.base_url
            if not url.endswith("/"):
                url += "/"
            self.control.setHtml( text , QtCore.QUrl.fromLocalFile ( url ) )
        else:
            self.control.setHtml( text )

    #-- Event Handlers ---------------------------------------------------------

    def _base_url_changed ( self ):
        self.update_editor()

    def _link_clicked ( self, url ):
        webbrowser.open_new( url.toString() )

#-EOF--------------------------------------------------------------------------
